##
## Filename    : chart_model.py
## Author(s)   : Michel Le Borgne
## Created     : 5/12/2011
## Revision    :
## Source      :
##
## Copyright 2009 - 2010  - 2011: IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
##     http:
##     mailto:
##
##
"""
A drawing style contains all the information to display a chart model.
We use essentially two styles Navigator and Plain corresponding to
the two views displayed on the gui.
"""
from math import sqrt

from gtk.gdk import color_parse
import pango


class DrawingStyle:
    """"
    Generic class for chart model drawing
    """

    def __init__(self):
        pass

    def draw_start(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        pass

    def draw_trap(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        pass


    def draw_top_node(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        pass

    def draw_macro(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        pass

    def draw_simple(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        pass

    def draw_perm(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        pass

    def draw_input(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        pass

    def draw_transition(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        pass

    def draw_transition_group(self, tgr, xfv, yfv, wfv, hfv):
        """
        Draw transitions with common extremities
        @param tgr: is a list of transitions with same extremities
        @param xfv,yfv: view coordinates of the macro node
        @param wfv,hfv: affinity ratios father node -> virtual screen
        """
        pass



        # visitors give virtual size of the node and limits
    def visit_cstart_node(self, node):
        """
        @return: (v_width, v_height, 0.0)
        """
        return (0.0, 0.0, 0.0)

    def visit_ctrap_node(self, node):
        """
        @return: (v_width, v_height, 0.0)
        """
        return (0.0, 0.0, 0.0)

    def visit_cinput_node(self, node):
        """
        @return: (v_width, v_height, 0.0)
        """
        return (0.0, 0.0, 0.0)

    def visit_csimple_node(self, node):
        """
        @return: (v_width, v_height, v_h_top_limit)
        """
        return (0.0, 0.0, 0.0)

    def visit_cperm_node(self, node):
        """
        @return: (v_width, v_height, v_h_top_limit)
        """
        return (0.0, 0.0, 0.0)


    def visit_cmacro_node(self, node):
        """
        @return: (v_width, v_height, v_h_top_limit, handle_w_limit, handle_h_limit)
        """
        return (0.0, 0.0, 0.0, 0.0, 0.0)

    def visit_ctop_node(self, node):
        """
        @return:(0.0, 0.0, 0.0)
        """
        return (0.0, 0.0, 0.0)

    def visit_ctransition(self, node):
        """
        @return: (v_w_detect_limit, v_h_detect_limit
        """
        return (0.0, 0.0)


# main drawing style

class PlainDrawing(DrawingStyle):
    """
    Main drawing style used in main window of charter
    """
    LABEL_H = 22.0
    W_SNODE = 65.0
    H_SNODE = 20.0
    W_INODE = 60.0
    H_INODE = 30.0
    U_SLOPE = H_SNODE/W_SNODE
    NRADIUS = 4
    SNODE_COLOR = "#EDFFEE"
    MNODE_COLOR = "ivory"
    ENODE_COLOR = "#f9fff9"
    PNODE_COLOR = "#fff7ed"
    INODE_COLOR = "green"

    SELECT_COLOR = "red"
    ACTIV_COLOR = "#FF9F7C"
    SEARCH_COLOR = "#DA70D6"

    DELTA = 7.0  # for handler detection
    DIST = 6.0 # distance for transition detection in pixel

    def __init__(self):
        self.view = None
        pass

    def draw_simple(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # coordinates in view
        xxr = int((xfv + node.xloc*wfv) * wview)
        yyr = int((yfv + node.yloc*hfv) * hview)
        # must be included in macro
#        wr = min(self.W_SNODE, (xfv - xr+wfv))
#        her = min(self.H_SNODE, (yfv*hview - yr+hfv))
        wir = self.W_SNODE
        her = self.H_SNODE
        # find colors
        if node.selected:
            line_color = self.SELECT_COLOR
        else:
            line_color = 'black'

        if node.activated:
            node_color = self.ACTIV_COLOR
        elif node.search_mark:
            node_color = self.SEARCH_COLOR
        else:
            node_color = self.SNODE_COLOR

        self.draw_node_gen(view, xxr, yyr, wir, her,
                           node.name, node_color, line_color)


    def draw_macro(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # coordinates in drawing window
        xxr = int((xfv + node.xloc*wfv) * wview)
        yyr = int((yfv + node.yloc*hfv) * hview)
        if node.model.show_macro:
            wir = int(node.wloc * wfv * wview)
            her = int(node.hloc * hfv * hview)
        else:
#            wir = min(self.W_SNODE, xfv - xr+wfv)
#            her = min(self.H_SNODE, yfv-yr+hfv)
            wir = self.W_SNODE
            her = self.H_SNODE
        if node.selected:
            line_color = self.SELECT_COLOR
        else:
            line_color = 'black'

        if node.model.show_macro:
            name = node.name
        else:
            name = node.name[0:9]

        if node.model.show_macro:
            self.draw_macro_node_gen(view, xxr, yyr, wir, her,
                                     self.LABEL_H, name,
                                     self.MNODE_COLOR, line_color)
        else:
            self.draw_node_gen(view, xxr, yyr, wir, her, name,
                               self.MNODE_COLOR, line_color)


    def draw_start(self, node, xfv, yfv, wfv, hfv):
        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # coordinates
        xxr = int((xfv + node.xloc*wfv)*wview)
        yyr = int((yfv + node.yloc*hfv)*hview + self.H_SNODE/2)
        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()
        grc.set_rgb_fg_color(color_parse('black'))
        wid = int(xxr-self.NRADIUS)
        hei = int(yyr-self.NRADIUS)
        pixmap.draw_arc(grc, True, wid, hei,
                        2 * self.NRADIUS, 2 * self.NRADIUS, 0, 360*64)

    def draw_trap(self, node, xfv, yfv, wfv, hfv):
        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # coordinates
        xxr = int((xfv + node.xloc*wfv)*wview)
        yyr = int((yfv + node.yloc*hfv)*hview)
        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()
        grc.set_rgb_fg_color(color_parse('black'))
        grc.line_width = 2
        wid = int(xxr-self.NRADIUS)
        hei = int(yyr-self.NRADIUS)
        pixmap.draw_arc(grc, False, wid, hei,
                        2 * self.NRADIUS, 2 * self.NRADIUS, 0, 360*64)

    def draw_perm(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """

        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # coordinates in view
        xxr = int((xfv + node.xloc*wfv)*wview)
        yyr = int((yfv + node.yloc*hfv)*hview)

        # must be included in macro
#        wir = min(self.W_SNODE, xfv - xxr+wfv)
#        her = min(self.H_SNODE, yfv-yyr+hfv)
        wir = self.W_SNODE
        her = self.H_SNODE
        # find colors
        if node.selected:
            line_color = self.SELECT_COLOR
        else:
            line_color = 'black'

        if node.activated:
            node_color = self.ACTIV_COLOR
        elif node.search_mark:
            node_color = self.SEARCH_COLOR
        else:
            node_color = self.PNODE_COLOR

        self.draw_node_gen(view, xxr, yyr, wir, her,
                           node.name, node_color, line_color)


    def draw_input(self, node, xfv, yfv, wfv, hfv):
        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # coordinates
        xxr = int((xfv + node.xloc*wfv)*wview + self.W_SNODE/2)
        yyr = int((yfv + node.yloc*hfv)*hview + self.H_SNODE/2)
        # color choice (activated or not)
        if node.activated:
            color = self.ACTIV_COLOR
        else:
            color = self.INODE_COLOR
        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()

        # diamond drawing
        rax = int(self.W_INODE/2.0)
        ray = int(self.H_INODE/2.0)

        grc.set_rgb_fg_color(color_parse(color))
        xr1 = xxr - rax
        xr2 = xxr + rax
        yr1 = yyr - ray
        yr2 = yyr + ray
        lpe = [(xr1, yyr), (xxr, yr1), (xr2, yyr), (xxr, yr2)]
        pixmap.draw_polygon(grc, True, lpe)
        if node.selected:
            color = self.SELECT_COLOR
        else:
            color = 'black'
        grc.set_rgb_fg_color(color_parse(color))
        grc.line_width = 2
        pixmap.draw_polygon(grc, False, lpe)
        #name
        view.stylepango.set_text(node.name)
        (xps, yps) = view.stylepango.get_pixel_size()
        xxx = xxr - xps/2.0
        yyy = yyr - yps/2.0
        pixmap.draw_layout(grc, int(xxx), int(yyy), view.stylepango)


    def draw_transition(self, trans, x1v, y1v, x2v, y2v, nnn, hor, xfv, yfv):
        """
        @param trans: the transition
        @param x1v, y1v: origin coordinates in virtual screen
        @param x2v, y2v: extremity coordinates in virtual screen
        @param xfv, yfv: father coordinate in virtual screen
        """
        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # find colors
        if trans.selected:
            color = self.SELECT_COLOR
        elif trans.search_mark:
            color = self.SEARCH_COLOR
        elif trans.activated:
            color = self.ACTIV_COLOR
        else:
            color = 'black'
        label = trans.event + "["+trans.condition+"]"+ trans.action

        # coordinates in view
        xr1 = int((x1v)*wview)
        yr1 = int((y1v)*hview)
        if trans.ori.is_start():
            yr1 = yr1 + int(self.H_SNODE/2)
        if trans.ori.is_input():
            xr1 = xr1 + int(self.W_SNODE/2)
            yr1 = yr1 + int(self.H_SNODE/2)
        xr2 = int((x2v)*wview)
        yr2 = int((y2v)*hview)
        self.draw_edge_gen(view, xr1, yr1, xr2, yr2, hor, nnn, label, color)

    def draw_transition_group(self, tgr, xfv, yfv, wfv, hfv):
        """
        Draw transitions with common extremities
        @param tgr: is a list of transitions with same extremities
        @param xfv, yfv: view coordinates of the macro node
        @param wfv, hfv: affinity ratios father node -> virtual screen
        """
        trans = tgr[0]
        ori = trans.ori
        ext = trans.ext
        nbtrans = len(tgr)

        # extremities coordinates in local frame
        (xo1, yo1, gap1, hor1) = ori.intersect(ext, self, nbtrans, wfv, hfv)
        (xe2, ye2, gap2, hor2) = ext.intersect(ori, self, nbtrans, wfv, hfv)

        # change to virtual view coordinates
        x1v = xfv + xo1 * wfv
        y1v = yfv + yo1 * hfv
        if hor1:
            gap1v = gap1*wfv
        else:
            gap1v = gap1*hfv

        x2v = xfv + xe2 * wfv
        y2v = yfv + ye2 * hfv
        if hor2:
            gap2v = gap2*wfv
        else:
            gap2v = gap2*hfv

        # draw each transition
        cpt = 1
        for trans in tgr:
            if trans.ori == ori: # same direction as first one
                self.draw_transition(trans, x1v, y1v, x2v, y2v,
                                     cpt, hor1, xfv, yfv)
                # record positions in virtual screen for find -
                # position in local coordinate
                trans.ori_coord = (xo1, yo1)
                trans.ext_coord = (xe2, ye2)
            else:  # opposite direction
                self.draw_transition(trans, x2v, y2v, x1v, y1v,
                                     cpt, hor1, xfv, yfv)
                # record positions in virtual screen for find
                # position in local frame
                trans.ext_coord = (xo1, yo1)
                trans.ori_coord = (xe2, ye2)
            # next position
            if hor1:
                x1v = x1v + gap1v
                xo1 = xo1 + gap1
            else:
                y1v = y1v + gap1v
                yo1 = yo1 + gap1
            if hor2:
                x2v = x2v + gap2v
                xe2 = xe2 + gap2
            else:
                y2v = y2v + gap2v
                ye2 = ye2 + gap2
            cpt += 1


    # visitor method
    # visitors give virtual size of the node and limits
    def visit_cstart_node(self, node):
        wid = float(self.view.draw_width)
        hei = float(self.view.draw_height)
        return (self.NRADIUS/wid, self.NRADIUS/hei, self.H_SNODE/(2 * hei))

    def visit_ctrap_node(self, node):
        wid = float(self.view.draw_width)
        hei = float(self.view.draw_height)
        return (self.NRADIUS/wid, self.NRADIUS/hei, 0.0)

    def visit_cinput_node(self, node):
        wid = float(self.view.draw_width)
        hei = float(self.view.draw_height)
        hlabel  = self.LABEL_H/hei
        return (self.W_INODE/wid, self.H_INODE/hei, hlabel)

    def visit_csimple_node(self, node):
        wid = float(self.view.draw_width)
        hei = float(self.view.draw_height)
        return (self.W_SNODE/wid, self.H_SNODE/hei, self.LABEL_H/hei)

    def visit_cperm_node(self, node):
        wid = float(self.view.draw_width)
        hei = float(self.view.draw_height)
        return (self.W_SNODE/wid, self.H_SNODE/hei, self.LABEL_H/hei)


    def visit_cmacro_node(self, node):
        wid = float(self.view.draw_width)
        hei = float(self.view.draw_height)
        del_x = self.DELTA/wid
        del_y = self.DELTA/hei
        return (self.W_SNODE/wid, self.H_SNODE/hei, self.LABEL_H/hei,
                del_x, del_y)

    def visit_ctop_node(self, node):
        return (0.0, 0.0, 0.0)

    def visit_ctransition(self, node):
        wid = float(self.view.draw_width)
        hei = float(self.view.draw_height)
        return (self.DELTA/wid, self.DELTA/hei)


    # generic drawing
    def draw_macro_node_gen(self, view, xxr, yyr, wir, her, hlabel, name,
                            color, line_color):
        """
        Rectangle with round corners
        """
        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()
        cac = pixmap.cairo_create()

        # rounded rectangle
        #   A****BQ
        #  H      C
        #  *      *
        #  G      D
        #   F****E
        col = color_parse(line_color)
        cac.set_line_width(1.0)

        ray = 12
        # Move to A
        cac.move_to(xxr + ray, yyr)
        # Straight line to B
        cac.line_to(xxr + wir - ray, yyr)
        # Curve to C, Control points are both at Q
        cac.curve_to(xxr + wir, yyr , xxr + wir, yyr, xxr + wir, yyr + ray)
        # Move to D
        cac.line_to(xxr + wir, yyr + her - ray)
        # Curve to E
        cac.curve_to(xxr + wir, yyr + her, xxr + wir, yyr + her,
                    xxr + wir - ray, yyr + her)
        # Line to F
        cac.line_to(xxr + ray, yyr + her)
        # Curve to G
        cac.curve_to(xxr, yyr + her, xxr, yyr + her, xxr, yyr + her - ray)
        # Line to H
        cac.line_to(xxr, yyr+ray)
        # Curve to A
        cac.curve_to(xxr, yyr, xxr, yyr, xxr + ray, yyr)

    #    cac.move_to(xr, yyr+LABEL_H)
    #    cac.line_to(xr+wir, yyr+LABEL_H)
        cac.set_source_rgba(col.red / 65535.0, col.green / 65535.0,
                            col.blue / 65535.0)
        cac.stroke_preserve()
        col = color_parse(color)
        cac.set_source_rgba(col.red / 65535.0, col.green / 65535.0,
                            col.blue / 65535.0)
        cac.fill()
        grc.set_rgb_fg_color(color_parse(line_color))
        pixmap.draw_line(grc, int(xxr), int(yyr + self.LABEL_H),
                         int(xxr + wir), int(yyr + self.LABEL_H))
        # label
        view.stylepango.set_text(name)
        fod = pango.FontDescription("normal 9")
        view.stylepango.set_font_description(fod)
        (xpl, ypl) = view.stylepango.get_pixel_size()
        # center
        xpos = xxr + wir/2 - xpl/2
        ypos = yyr + 4
        pixmap.draw_layout(grc, int(xpos), int(ypos), view.stylepango)


    def draw_node_gen(self, view, xxr, yyr, wir, her, name, color, line_color):
        """
        @param xxr,yyr: coordinates in view
        @param wir, her: size in view
        """
        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()
        cac = pixmap.cairo_create()
        # rounded rectangle
        #   A****BQ
        #  H      C
        #  *      *
        #  G      D
        #   F****E
        col = color_parse(line_color)
        cac.set_line_width(1.5)

        ray = 7
        # Move to A
        cac.move_to(xxr + ray, yyr)
        # Straight line to B
        cac.line_to(xxr + wir - ray, yyr)
        # Curve to C, Control points are both at Q
        cac.curve_to(xxr + wir, yyr, xxr + wir, yyr, xxr + wir, yyr + ray)
        # Move to D
        cac.line_to(xxr + wir, yyr + her - ray)
        # Curve to E
        cac.curve_to(xxr + wir, yyr + her, xxr + wir, yyr + her,
                     xxr + wir - ray, yyr + her)
        # Line to F
        cac.line_to(xxr + ray, yyr + her)
        # Curve to G
        cac.curve_to(xxr, yyr + her, xxr, yyr + her, xxr, yyr + her - ray)
        # Line to H
        cac.line_to(xxr, yyr + ray)
        # Curve to A
        cac.curve_to(xxr, yyr, xxr, yyr, xxr + ray, yyr)
        cac.set_source_rgba(col.red / 65535.0, col.green / 65535.0,
                            col.blue / 65535.0)
        cac.stroke_preserve()
        col = color_parse(color)
        cac.set_source_rgba(col.red / 65535.0, col.green / 65535.0,
                            col.blue / 65535.0)
        cac.fill()

        # label
        view.stylepango.set_text(name[:9])
        fod  = pango.FontDescription("normal 9")
        view.stylepango.set_font_description(fod)
        (xpl, ypl)=view.stylepango.get_pixel_size()
        # center
        xpos = xxr + wir/2 - xpl/2
        ypos = yyr + 4
        pixmap.draw_layout(grc, int(xpos), int(ypos), view.stylepango)

    def draw_edge_gen(self, view, xr1, yr1, xr2, yr2, hor, nnn, label, color):
        """
        draw an edge
        """
        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()
        grc.set_rgb_fg_color(color_parse(color))
        # line
        pixmap.draw_line(grc, xr1, yr1, xr2, yr2)
        # arrow
        vux = xr2 - xr1
        vuy = yr2 - yr1
        norm = sqrt(vux ** 2 + vuy ** 2)
        if norm != 0:
            vux = vux / norm
            vuy = vuy / norm
            arr = Arrow()
            arr.draw(view, (vux, vuy),(xr2, yr2))
        # label
        if len(label)>20:
            label1 = label[0:18]+'...'
        else:
            label1 = label
        old_fd = view.stylepango.get_font_description()
        fod = pango.FontDescription("normal 9")
        view.stylepango.set_font_description(fod)
        view.stylepango.set_text(label1)
        (wpl, hpl) = view.stylepango.get_pixel_size()
        # draw label if enough room
        if wpl < abs(xr2 - xr1) or 2 * hpl < abs(yr2 - yr1):
            if not hor:
                xxl = (xr1 + xr2) / 2 - wpl / 2
                yyl = (yr1 + yr2) / 2 - hpl + 2
                pixmap.draw_layout(grc, xxl, yyl, view.stylepango)
            else:
                xxl = (nnn * xr1 + xr2) / (nnn + 1) - wpl / 2
                yyl = (nnn * yr1 + yr2) / (nnn + 1)
                pixmap.draw_layout(grc, xxl, yyl, view.stylepango)
        view.stylepango.set_font_description(old_fd)


class NavDrawing(DrawingStyle):
    """
    Drawing style for an overview
    """
    W_SNODE = 8.0
    H_SNODE = 6.0
    W_INODE = 6.0
    H_INODE = 3.0
    U_SLOPE = H_SNODE/W_SNODE
    NRADIUS = 3
    SNODE_COLOR = "#EDFFEE"
    MNODE_COLOR = "ivory"
    ENODE_COLOR = "#f9fff9"
    PNODE_COLOR = "#FFC1C1"          #"#ffabab"
    INODE_COLOR = "green"
    TNODE_COLOR = 'black'
    StNODE_COLOR = 'blue'

    SELECT_COLOR = "red"
    ACTIV_COLOR = "#FF9F7C"
    SEARCH_COLOR = "#CB1DC4"

    DELTA = 7.0  # for handler detection
    DIST = 6.0 # distance for transition detection in pixel
    def __init__(self):
        self.view = None
        pass


    def draw_top_node(self, node, xfv, yfv, wfv, hfv):
        pass

    def draw_macro(self, node, xfv, yfv, wfv, hfv):
        """
        @param xfv: x coordinate of father in virtual view
        @param yfv: y coordinate of father in virtual view
        @param wfv: width of father in virtual view
        @param hfv: height of father in virtual view
        """
        if not node.model.show_macro:
            self.draw_rectangle(node, xfv, yfv, wfv, hfv, True,
                                self.MNODE_COLOR)
            return
        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # coordinates
        xxr = int((xfv + node.xloc * wfv) * wview)
        yyr = int((yfv + node.yloc * hfv) * hview)
        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()
        grc.set_rgb_fg_color(color_parse(self.MNODE_COLOR))
        wid = int(node.wloc * wfv * wview)
        hei = int(node.hloc * hfv * hview)
        pixmap.draw_rectangle(grc, True, xxr, yyr, wid, hei)
        grc.set_rgb_fg_color(color_parse('black'))
        pixmap.draw_rectangle(grc, False, xxr, yyr, wid, hei)

    def draw_start(self, node, xfv, yfv, wfv, hfv):
        self.draw_circle(node, xfv, yfv, wfv, hfv, True, self.StNODE_COLOR)

    def draw_trap(self, node, xfv, yfv, wfv, hfv):
        self.draw_circle(node, xfv, yfv, wfv, hfv, True, self.TNODE_COLOR)

    def draw_simple(self, node, xfv, yfv, wfv, hfv):
        self.draw_rectangle(node, xfv, yfv, wfv, hfv, True, self.SNODE_COLOR)

    def draw_perm(self, node, xfv, yfv, wfv, hfv):
        self.draw_rectangle(node, xfv, yfv, wfv, hfv, True, self.PNODE_COLOR)

    def draw_input(self, node, xfv, yfv, wfv, hfv):
        self.draw_circle(node, xfv, yfv, wfv, hfv, True, self.INODE_COLOR)

    def draw_transition(self, node, xfv, yfv, wfv, hfv):
        pass

    def draw_transition_group(self, tgr, xfv, yfv, wfv, hfv):
        """
        Draw transitions with common extremities
        @param tgr: is a list of transitions with same extremities
        @param xfv,yfv: view coordinates of the macro node
        @param wfv,hfv: affinity ratios father node -> virtual screen
        """
        trans = tgr[0]
        ori = trans.ori
        ext = trans.ext

        # change to virtual view coordinates
        x1v = xfv + ori.xloc * wfv
        y1v = yfv + ori.yloc * hfv
        x2v = xfv + ext.xloc * wfv
        y2v = yfv + ext.yloc * hfv

        view = self.view
        v_size = ori.accept(self)
        if ori.is_macro():
            if ori.model.show_macro:
                xcr = v_size[0] * wfv * view.draw_width
                ycr = v_size[1] * hfv * view.draw_height
            else:
                xcr = (self.W_SNODE / 2.0) / view.draw_width
                ycr = (self.H_SNODE / 2.0) / view.draw_height
        else:
            xcr = v_size[0]
            ycr = v_size[1]
        x1r = int(x1v * view.draw_width + xcr)
        y1r = int(y1v * view.draw_height + ycr)

        v_size = ext.accept(self)
        if ext.is_macro():
            if ext.model.show_macro:
                xcr = v_size[0] * wfv * view.draw_width
                ycr = v_size[1] * hfv * view.draw_height
            else:
                xcr = (self.W_SNODE/2.0)/view.draw_width
                ycr = (self.H_SNODE/2.0)/view.draw_height
        else:
            xcr = v_size[0]
            ycr = v_size[1]
        x2r = int(x2v * view.draw_width + xcr)
        y2r = int(y2v * view.draw_height + ycr)
        pixmap = view.pixmap
        grc = view.window.new_gc()
        grc.set_rgb_fg_color(color_parse('black'))
        # line
        pixmap.draw_line(grc, x1r, y1r, x2r, y2r)


    def draw_circle(self, node, xfv, yfv, wfv, hfv, fill, color):
        """
        As it says
        """
        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # coordinates
        xxr = int((xfv + node.xloc * wfv) * wview)
        yyr = int((yfv + node.yloc * hfv) * hview)
        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()
        grc.set_rgb_fg_color(color_parse(color))
        wid = int(xxr - self.NRADIUS)
        hei = int(yyr - self.NRADIUS)
        pixmap.draw_arc(grc, True, wid, hei,
                        2 * self.NRADIUS, 2 * self.NRADIUS, 0, 360 * 64)
        grc.set_rgb_fg_color(color_parse('black'))
        pixmap.draw_arc(grc, False, wid, hei, 2 * self.NRADIUS,
                        2 * self.NRADIUS, 0, 360 * 64)

    def draw_rectangle(self, node, xfv, yfv, wfv, hfv, fill, color):
        """
        As it says
        """
        view = self.view
        wview = view.draw_width
        hview = view.draw_height
        # coordinates
        xxr = int((xfv + node.xloc*wfv)*wview)
        yyr = int((yfv + node.yloc*hfv)*hview)
        # color
        if node.activated:
            color = self.ACTIV_COLOR
        elif node.search_mark:
            color = self.SEARCH_COLOR
        # graphic context
        pixmap = view.pixmap
        grc = view.window.new_gc()
        grc.set_rgb_fg_color(color_parse(color))
        wid = int(self.W_SNODE)
        hei = int(self.H_SNODE)
        pixmap.draw_rectangle(grc, True, xxr, yyr, wid, hei)
        grc.set_rgb_fg_color(color_parse('black'))
        pixmap.draw_rectangle(grc, False, xxr, yyr, wid, hei)

    def visit_cstart_node(self, node):
        """
        @return: (v_width, v_height, 0.0)
        """
        return (self.NRADIUS/2, self.NRADIUS/2, 0.0)

    def visit_ctrap_node(self, node):
        """
        @return: (v_width, v_height, 0.0)
        """
        return (self.NRADIUS/2.0, self.NRADIUS/2.0, 0.0)

    def visit_cinput_node(self, node):
        """
        @return: (v_width, v_height, 0.0)
        """
        return (0.0, 0.0, 0.0)

    def visit_csimple_node(self, node):
        """
        @return: (v_width, v_height, v_h_top_limit)
        """
        return (self.W_SNODE/2, self.H_SNODE/2, 0.0)

    def visit_cperm_node(self, node):
        """
        @return: (v_width, v_height, v_h_top_limit)
        """
        return (self.W_SNODE/2, self.H_SNODE/2, 0.0)


    def visit_cmacro_node(self, node):
        """
        @return: (v_width, v_height, v_h_top_limit, handle_w_limit, handle_h_limit)
        """
        return (node.wloc/2.0, node.hloc/2.0, 0.0, 0.0, 0.0)

    def visit_ctop_node(self, node):
        """
        @return:(0.0, 0.0, 0.0)
        """
        return (0.0, 0.0, 0.0)

    def visit_ctransition(self, node):
        """
        @return: (v_w_detect_limit, v_h_detect_limit
        """
        return (0.0, 0.0)





# arrow
ALENGTH = 10
SAANGLE = 0.342020
CAANGLE = 0.939693
ARROWCOLOR = "black"


class Arrow:
    """
    Generic arrow
    """
    def __init__(self):
        pass

    def draw(self, view, direct, pos):
        """
        draw the arrow
        """
        pixmap = view.pixmap
        vux = - direct[0]
        vuy = - direct[1]
        poly = [(int(pos[0]), int(pos[1]))]
        xx0 = pos[0]
        yy0 = pos[1]
        u1x = vux * CAANGLE - vuy * SAANGLE
        u1y = vux * SAANGLE + vuy * CAANGLE
        xxx = int(xx0 + ALENGTH * 1.2 * u1x)
        yyy = int(yy0 + ALENGTH * 1.2 * u1y)
        poly.append((xxx, yyy))
        xxx = int(xx0 + ALENGTH * vux)
        yyy = int(yy0 + ALENGTH * vuy)
        poly.append((xxx, yyy))
        u1x = vux * CAANGLE + vuy * SAANGLE
        u1y = vuy * CAANGLE - vux * SAANGLE
        xxx = int(xx0 + ALENGTH * 1.2 * u1x)
        yyy = int(yy0 + ALENGTH * 1.2 * u1y)
        poly.append((xxx, yyy))
        grc = view.window.new_gc()
        grc.set_rgb_fg_color(color_parse(ARROWCOLOR))
        pixmap.draw_polygon(grc, True, poly)



## Filename    : layout
## Author(s)   : Michel Le Borgne
## Created     : 01/2012
## Revision    : 
## Source      : 
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA/IRSET
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE 
##     
##
## Contributor(s):
##
"""
Layouts are performed by a visitor of the chart model
"""
import sys
from string import atof
import pygraphviz as pgv

class LayoutVisitor(object):
    """
    Visitor used to do a gt graph layout
    """
    def __init__(self, view, layout_style):
        self.view = view
        self.drawing_style = view.drawing_style
        self.layout = layout_style
        
    
    def visit_chart_model(self, model):
        """
         Not used
        """
        model.get_root().accept(self)
        
    def visit_cstart_node(self, node):
        """
         Not used
        """
        return
    
    def visit_ctrap_node(self, node):
        """
         Not used
        """
        return
    
    def visit_csimple_node(self, node):
        """
         Not used
        """
        return
    
    def visit_cinput_node(self, node):
        """
         Not used
        """
        return

    def visit_cperm_node(self, node):
        """
         Not used
        """
        return
    
    def visit_cmacro_node(self, node):
        """
        Layouts are done on each macro place
        Change the local coordinates of subnodes according to layout style 
        """
        if node.sub_nodes == []:
            return
        agraph = pgv.AGraph(splines='ortho')
        node_dict = dict()
        # build nodes
        for snode in node.sub_nodes:            
            agraph.add_node(snode.name)
            anode = agraph.get_node(snode.name)
            node_dict[snode.name] = snode
            if snode.is_macro() and snode.model.show_macro:
                anode.attr['width'] = "%s" % snode.wloc
                anode.attr['height'] = "%s" % snode.hloc
                snode.accept(self)
        # build edges
        for tgr in node.transitions:
            for trans in tgr:
                agraph.add_edge(trans.ori.name, trans.ext.name)
        # layout
        layout_style = self.layout
        if layout_style == 'hierarchical_LR':
            agraph.layout(prog='dot', args='-Grankdir=LR')
        elif layout_style == 'hierarchical_TB':
            agraph.layout(prog='dot', args='-Grankdir=BT')
        elif layout_style == 'neato':
            agraph.layout(prog='neato')
        elif layout_style == 'fdp':
            agraph.layout(prog='fdp')
        elif layout_style == 'twopi':
            agraph.layout(prog='twopi')
        elif layout_style == 'circo':
            agraph.layout(prog='circo')
        else:
            print 'BUG IN MACRO LAYOUT!!'
            sys.exit(1)
        # change coordinates
        # bounding box for ratios
        bbox = pgv.graphviz.agget(agraph.handle, 'bb')
        bbox = bbox.split(',')
        xmin = atof(bbox[0])
        xmax = atof(bbox[2])
        ymin = atof(bbox[1])
        ymax = atof(bbox[3])
        #adjust graphic window
        view = self.view
        cond = ((xmax - xmin) / 3.0 > view.draw_width) 
        cond = cond or ((xmax - xmin) / 3.0 < view.draw_width / 2.0) 
        if cond:
            view.draw_width = min(int((xmax - xmin)/3.0), 10400) # 31200 max
            view.draw_width = max(1300, view.draw_width)
            view.min_width = view.draw_width
        cond = ymax - ymin > view.draw_height
        cond = cond  or ymax - ymin < view.draw_height/2.0
        if cond: 
            view.draw_height = min(int(ymax - ymin), 7200) # 21600 max
            view.draw_height = max(900, view.draw_height)
            view.min_height = view.draw_height
        # reset zoom counter
        view.zoom_count = 0
        gwidth = 1.0
        gheight = 1.0 # virtual window for the macro node
        if xmin == xmax or ymin == ymax:
            # problem TODO
            sys.exit(1)
            pass
        x_ratio = gwidth/(xmax-xmin)
        y_ratio = gheight/(ymax-ymin)
        # changing coordinates
        for node in agraph.nodes():
            pos = node.attr['pos'].split(',')
            cnn = node_dict[node]
            xcoord = atof(pos[0]) * x_ratio
            ycoord = atof(pos[1]) * y_ratio
            cnn.set_layout_coordinates(xcoord, ycoord)
        return

    def visit_ctop_node(self, node):
        """
        Same as macro node
        """
        self.visit_cmacro_node(node)
        return
    
    def visit_ctransition(self, trans):
        """
         Not used
        """
        return
    



##
## Filename    : reporter.py
## Author(s)   : Michel Le Borgne
## Created     : 10/2009
## Revision    : 
## Source      : 
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE 
##     
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##

"""
To adapt message handling to different environment we use reporters
"""

# simple reporter
class SimpleErrorReporter(object):
    """
    Minimum reporter
    """
    def __init__(self):
        self.error = False
        
    
    def display(self, mess):
        """
        Set the error attribute and print the message
        """
        self.error = True
        print '\n>> '+mess



class ErrorReporter(object):
    """
    Usefull for some tests where a header message is informative
    display messages in a textviewer
    """
    def __init__(self, tview, model):
        self.display_window = tview # must be a textviewer
        self.mod_identifier = model
        self.text = ""
        self.err_msg = False
        
    def display(self, hdr, mes = ""):
        """
        build the message from mess and hdr
        """
        buf = self.display_window.get_buffer()
        iter_end = buf.get_end_iter()
        msg = self.mod_identifier+ ':: '+ hdr +'  ' + mes + '\n'
        buf.insert(iter_end, msg)
        self.text = msg
        self.err_msg = True
        
class CompilReporter(object):
    """
    register errors and info - must be checked after compilation
    memory contains all the messages
    """
    def __init__(self):
        self.error = False
        self.info = False
        self.use_context = True
        self.context = ""
        self.memory = ""
        self.indent = "       "
    
    def set_context(self, mess):
        """
        Context message: describe the context (transition for exp.)
        """
        self.context = mess
        self.use_context = False
        
    def display(self, mess, supl = ""):
        """
        Combine the message with context information
        """
        if not self.use_context:
            self.memory = self.memory+'\n\n'+self.context
            self.use_context = True
        self.error = True
        self.memory = self.memory+'\n'+ self.indent + mess +'  '+supl
        
    def display_info(self, mess, supl = ""):
        """
        For information
        """
        if not self.use_context:
            self.memory = self.memory+'\n\n'+self.context
            self.use_context = True
        self.info = True
        self.memory = self.memory+'\n'+ self.indent + mess +'  '+supl   

    def reset(self):
        """
        RAZ
        """
        self.error = False
        self.info = False
        self.use_context = True
        self.context = ""
        self.memory = ""





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2010-2017  IRISA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# The original code contained here was initially developed by:
#
#     Geoffroy Andrieux.
#     IRISA/IRSET
#     Symbiose team
#     IRISA  Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
#
#
# Contributor(s): Michel Le Borgne, Nolwenn Le Meur, Pierre Vignet

"""Definition of setup function for setuptools module."""

# Standard imports
from setuptools import setup, find_packages

__PACKAGE_VERSION__ = "0.2.2"
__LIBRARY_VERSION__ = "0.1.0"

setup(

    # Library name & version
    name='cadbiom_gui',
    version=__PACKAGE_VERSION__,

    # Search all packages recursively
    packages=find_packages(),

    # Authors
    author="pvignet",
    author_email="pierre.vignet@irisa.fr",

    # Description
    description="GUI tool for the Cadbiom library",
    long_description=open('README.md').read(),

    # MANIFEST.in ?
    include_package_data=True,

    # Official page
    url = "https://gitlab.irisa.fr/0000B8EG/Cadbiom/tree/pypi_packaging",

    # Metadata
    classifiers=[
        "Environment :: X11 Applications :: GTK",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Natural Language :: French",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],

    entry_points={
        'gui_scripts': ['cadbiom = cadbiom_gui.__main__:main'],
    },

    install_requires=["cadbiom", "pygraphviz"],
)



## Filename    : charter.py
## Author(s)   : Geoffroy Andrieux
## Created     : 03/2010
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Michel Le Borgne
##
"""
Small widgets for Cadbiom gui
"""
import gtk
import re
import traceback
from utils.listDisplay import ToggleList
from cadbiom.models.guard_transitions.analyser.ana_visitors import FrontierVisitor
from utils.warn import cancel_warn
from cadbiom.models.guard_transitions.chart_model import ChartModel
from cadbiom.models.guard_transitions.analyser import model_corrections
from cadbiom.models.guard_transitions.translators.chart_xml import \
    MakeModelFromXmlFile

from utils.fileHandling import FileChooser
from utils.reporter import CompilReporter
from cadbiom.models.guard_transitions.translators.chart_xml_pid import \
            MakeModelFromPidFile
from chart_simulator.chart_simul_controler import DisplayError

import pkg_resources

class SearchManager(object):
    """
    List of simple nodes for searching
    """
    def __init__(self, chart, notebook, label):
        self.charter = chart
        self.model = None
        self.model_changed = True
        self.notebook = notebook

        # simple nodes
        sn_frame = gtk.Frame()
        self.sn_frame = sn_frame
        vbox = gtk.VBox(False, 0)
        sn_frame.add(vbox)
        # simple nodes list
        self.sn_viewer = ToggleList()
        label = gtk.Label(label)
        # wrap into scrollwindow
        scroll = gtk.ScrolledWindow()
        scroll.add_with_viewport(self.sn_viewer)
        vbox.pack_start(scroll, True, True, 0)
        # buttons
        hbox = gtk.HBox()
        vbox.pack_start(hbox, False, False, 0)
        but = gtk.Button('Update')
        but.connect("clicked", self.on_update)
        self.show_but = but
        hbox.pack_start(but, False, False, 0)
        but = gtk.Button('Show')
        but.connect("clicked", self.on_show)
        self.show_but = but
        hbox.pack_start(but, False, False, 0)
        but = gtk.Button('Extract')
        but.connect("clicked", self.on_extract)
        self.show_but = but
        hbox.pack_start(but, False, False, 0)
        but = gtk.Button('Clear')
        but.connect("clicked", self.on_clear)
        self.clear_but = but
        hbox.pack_start(but, False, False, 0)
        notebook.append_page(sn_frame, label)
        # entry
        self.entry = gtk.Entry()
        vbox.pack_start(self.entry, False, False, 0)

    def update(self):
        """
        As it says
        """
        self.model_changed = True

    def set_model(self, model):
        """
        As it says
        """
        if self.model:
            self.model.detach(self)
        self.model = model
        self.model.attach(self)
        self.model_changed = False
        # get node names
        lnode = model.get_simple_node_names()
        lnode.sort()
        # display node names
        self.sn_viewer.refresh(lnode)

    def on_clear(self, widget):
        """
        clear button callback
        """
        self.sn_viewer.clear(widget)
        self.entry.set_text('')
        if self.model:
            self.model.search_unmark()
            self.model_changed = False # search_mark implies a notify

    def on_update(self, widget):
        """
        update button callback
        """
        if self.model:
            self.model.search_unmark()
            # get node names
            lnode = self.model.get_simple_node_names()
            lnode.sort()
            # display node names
            self.sn_viewer.refresh(lnode)
            self.model_changed = False

    def on_show(self, widget):
        """
        show button callback
        """
        if not self.model:
            return
        # check model
        if self.model_changed:
            cancel_warn("Model is modified: update?")
        # mark nodes and show
        lnode = self.sn_viewer.get_selected_items()
        if lnode == []:
            regex = self.entry.get_text()
            if len(regex) == 0:
                return
            try:
                regex_obj = re.compile(regex)
            except:
                cancel_warn("Incorrect regular expression")
                return
            lnode = self.model.get_matching_node_names(regex_obj)

        self.model.search_mark(lnode)
        self.model_changed = False # search_mark implies a notify

    def on_extract(self, widget):
        """
        extract node environment
        """
        if not self.model:
            return
        # check model
        if self.model_changed:
            cancel_warn("Model is modified: update?")
        # mark nodes and show
        lnode_n = self.sn_viewer.get_selected_items()
        if lnode_n == []:
            regex = self.entry.get_text()
            if len(regex)==0:
                return
            try:
                regex_obj = re.compile(regex)
            except:
                cancel_warn("Incorrect regular expression")
                return
            lnode_n = self.model.get_matching_node_names(regex_obj)

        submodel = ChartModel(self.model.name + "_extract")
        tnode = submodel.get_root()
        for node_n in lnode_n:
            # avoid double node declaration
            try:
                extn = submodel.node_dict[node_n]
                break
            except:
                node = self.model.node_dict[node_n]
                nsn = tnode.add_copy(node)
            extn = self.model.node_dict[node_n]
            # incoming and out going transitions except with start or trap
            for trans in extn.incoming_trans:
                ori_n = trans.ori.name
                try:
                    ori = submodel.node_dict[ori_n]
                except:
                    ori = tnode.add_copy(trans.ori)
                ntr = tnode.add_transition(ori, nsn)
                if ntr:
                    ntr.set_condition(trans.condition)
                    ntr.set_event(trans.event)
                    ntr.set_action(trans.action)
            for trans in extn.outgoing_trans:
                ext_n = trans.ext.name
                try:
                    target = submodel.node_dict[ext_n]
                except:
                    target = tnode.add_copy(trans.ext)
                ntr = tnode.add_transition(nsn, target)
                if ntr:
                    ntr.set_condition(trans.condition)
                    ntr.set_event(trans.event)
                    ntr.set_action(trans.action)
            # transitions conditioned by node
            try:
                trl = self.model.signal_transition_dict[node]
            except:
                trl = []
            for trans in trl:
                ori_n = trans.ori.name
                try:
                    ori = submodel.node_dict[ori_n]
                except:
                    ori = tnode.add_copy(trans.ori)
                ext_n = trans.ext.name
                try:
                    target = submodel.node_dict[ext_n]
                except:
                    target = tnode.add_copy(trans.ext)
                ntr = tnode.add_transition(ori, target)
                if ntr:
                    ntr.set_condition(trans.condition)
                    ntr.set_event(trans.event)
                    ntr.set_action(trans.action)
        # display in charter
        self.charter.add_edit_mvc(submodel.name, submodel, False)
        self.charter.do_layout(None, "hierarchical_LR")

    def get_selected_keywords(self):
        """
        As it says
        """
        return self.sn_viewer.get_selected_items()

    def clear(self):
        """
        It is clear
        """
        self.sn_viewer.refresh([])

class SearchFrontier(SearchManager):
    """
    Same as SearchManager but specialized on frontier nodes
    """
    def __init__(self, chart, notebook, label):
        SearchManager.__init__(self, chart, notebook, label)

    def set_model(self, model):
        """
        As it says
        """
        if self.model:
            self.model.detach(self)
        self.model = model
        self.model.attach(self)
        self.model_changed = False
        # get node names
        lnode = self.get_frontier_node_names()
        lnode.sort()
        # display node names
        self.sn_viewer.refresh(lnode)

    def on_update(self, widget):
        self.model.search_unmark()
        # get node names
        lnode = self.get_frontier_node_names()
        lnode.sort()
        # display node names
        self.sn_viewer.refresh(lnode)
        self.model_changed = False

    def get_frontier_node_names(self):
        """
        As it says
        """
        fvi = FrontierVisitor()
        self.model.accept(fvi)
        return fvi.frontier

    def clear(self):
        self.sn_viewer.refresh([])


class LegendWindow(object):
    """
    Widget to display the legend
    """
    def __init__(self, parent=None):
        self.win = gtk.Window()
        self.win.set_title("CADBIOM-Chart Legend")
        self.win.set_position(gtk.WIN_POS_CENTER)
        self.win.connect("destroy", self.on_destroy)
        self.win.connect('key_press_event', self.on_escape)
        image = gtk.Image()
        template = pkg_resources.resource_filename(
            __name__,
            "images/legend.png"
        )
        image.set_from_file(template)
        image.show()

        self.win.add(image)
        self.win.show_all()
        # register itself for parent or emvc
        if parent:
            parent.win_register(self)

        self.parent = parent

    def on_destroy(self, widget):
        """
        standard destroy callback
        """
        if self.parent:
            self.parent.win_remove(self)

    def destroy(self):
        """
        if registered as a child
        """
        if self.win:
            self.win.destroy()

    def on_escape(self, widget, event):
        """On ESC key_press_event, destroy this window."""
        if gtk.gdk.keyval_name(event.keyval) == "Escape":
            self.destroy()


class ImportPIDParam(object):
    """
    Widget for model importing
    """
    def __init__(self, chart, parent):
        self.charter = chart

        # window creation
        template = pkg_resources.resource_filename(
            __name__,
            "chart_glade/import_parameter.glade"
        )
        self.wtree = gtk.glade.XML(template)
        self.main_window = self.wtree.get_widget("window1")
        self.main_window.set_title("Import window")

        self.main_window.set_resizable(True)
        hei = gtk.gdk.screen_height()
        hei = int(hei * 0.20)
        self.main_window.set_size_request(350, hei)

        # Set modal mode for the window (above all windows & block inputs)
        self.main_window.set_modal(True)
        self.main_window.set_transient_for(parent)

        if self.main_window:
            self.main_window.connect("destroy", self.on_destroy)
            self.main_window.connect('key_press_event', self.on_escape)

        self.main_window.set_position(gtk.WIN_POS_CENTER)

        # init param
        self.ai_inter = 0
        self.has_clock = False

        # interpretation radio button
        rbut = self.wtree.get_widget("or_rb")
        rb_name = rbut.get_name()
        rbut.connect("toggled", self.ai_inter_rb_callback, rb_name)

        rbut = self.wtree.get_widget("and_rb")
        rb_name = rbut.get_name()
        rbut.connect("toggled", self.ai_inter_rb_callback, rb_name)

        # Uncomment to add new buttons
#        rb = self.wtree.get_widget("or_and_rb")
#        rb_name = rb.get_name()
#        rb.connect("toggled", self.ai_inter_rb_callback, rb_name)
#
#        rb = self.wtree.get_widget("and_or_rb")
#        rb_name = rb.get_name()
#        rb.connect("toggled", self.ai_inter_rb_callback, rb_name)

        # clock radio button
        rbut = self.wtree.get_widget("withoutClock_rb")
        rbut.connect("toggled", self.clock_rb_callback, False)

        rbut = self.wtree.get_widget("withClock_rb")
        rbut.connect("toggled", self.clock_rb_callback, True)

        # import button
        button = self.wtree.get_widget("importButton")
        button.connect("clicked", self.on_import)

        # display
        self.main_window.show_all()

    def ai_inter_rb_callback(self, widget, ai_name):
        """
        set activator/inhibitor interpretation (and or or)
        """
        if ai_name == "or_rb":
            self.ai_inter = 0
        elif ai_name == "and_rb":
            self.ai_inter = 1

        # Uncomment to add new buttons
#        elif ai_name == "or_and_rb":
#            self.ai_inter = 2
#        else :
#            self.ai_inter = 3

    def clock_rb_callback(self, widget, has_clock):
        """
        set clock generation
        """
        if has_clock :
            self.has_clock = True
        else :
            self.has_clock = False

    def on_import(self, widget):
        """
        lauch import
        """
        fch = FileChooser("Import from PID xml", "xml files", "*.xml",
                          save_window=False)
        fch.do_action(self.import_from_pid_file)

    def import_from_pid_file(self, file):
        """
        compile a pid file
        """
        crep = CompilReporter()
        parser = MakeModelFromPidFile(file, crep,
                                          self.has_clock, self.ai_inter)
        if crep.error :
            DisplayError(crep, parser.model.name)
        else :
            model = parser.model
            model.modified = False
            self.charter.add_edit_mvc(model.name, model)
            self.destroy()

    def on_destroy(self, widget):
        """
        standard destroy callback
        """
        if self.main_window:
            self.main_window.destroy()

    def destroy(self):
        """
        for parents
        """
        self.on_destroy(None)

    def on_escape(self, widget, event):
        """On ESC key_press_event, destroy this window."""
        if gtk.gdk.keyval_name(event.keyval) == "Escape":
            self.destroy()


class ImportBioPAXParams(object):
    """
    Widget for BioPAX model importing

    .. note:: requires biopax2cadbiom module
    """
    def __init__(self, chart, parent):
        self.charter = chart

        try:
            from biopax2cadbiom.commons import SPARQL_PATH
        except ImportError:
            d_err = DisplayError(
                None, "biopax2cadbiom module is missing.",
                "biopax2cadbiom project is required to perform this task.\n" + \
                "Please install it via the usual command:\n" + \
                "pip install biopax2cadbiom"
            )
            d_err.main_window.set_modal(True)
            d_err.main_window.set_transient_for(parent)
            return

        # Init class variables (settings of the window)
        self.cadbiom_file = ''
        self.convert_full_graph = True
        self.graph_uris = list()
        self.full_compartments_names = True
        self.blacklist_file = None
        self.triplestore_url = SPARQL_PATH
        self.remove_strongly_connected_components = True

        # window creation
        template = pkg_resources.resource_filename(
            __name__,
            "chart_glade/import_BioPAX_parameters.glade"
        )
        self.wtree = gtk.glade.XML(template)
        self.main_window = self.wtree.get_widget("window1")
        self.main_window.set_title("Import window")
        self.main_window.set_resizable(True)
        self.main_window.set_position(gtk.WIN_POS_CENTER)

        # Set modal mode for the window (above all windows & block inputs)
        self.main_window.set_modal(True)
        self.main_window.set_transient_for(parent)

        if self.main_window:
            self.main_window.connect("destroy", self.on_destroy)
            self.main_window.connect('key_press_event', self.on_escape)

        # Init interface
        # Triplestore
        # Graphs URIs
        self.wtree.get_widget('text_graphs_uris').set_text(
            "http://reactome.org/mycobacterium"
        )
        # Triplestore URL
        self.wtree.get_widget('text_triplestore_url').set_text(
            self.triplestore_url)
        # BioPAX level
        self.wtree.get_widget('radio_biopax_level3').set_active(True)
        # BioPAX options
        self.wtree.get_widget('checkbox_convert_full_graph').set_active(False)
        self.wtree.get_widget('checkbox_full_compartment_names').set_active(True)
        self.wtree.get_widget('checkbox_remove_scc').set_active(True)

        # Connect buttons
        # Files
        self.wtree.get_widget("button_blacklist").connect(
            "clicked", self.on_import_blacklist_file
        )
        self.wtree.get_widget("button_output").connect(
            "clicked", self.on_set_output_file
        )
        self.wtree.get_widget("button_make_model").connect(
            "clicked", self.import_BioPAX_data
        )

        # Display
        self.main_window.show_all()


    def on_import_blacklist_file(self, widget):
        """
        Choose blacklist file of entities
        """
        fch = FileChooser("Choose the blacklist file", "txt files", "*.txt",
                          save_window=False)
        fch.do_action(self.set_blacklist_file)


    def on_set_output_file(self, widget):
        """
        Choose output model file
        """
        fch = FileChooser("Choose the output file name", "bcx files", "*.bcx")
        fch.do_action(self.set_output_file)
        # Change color to green
        self.wtree.get_widget("button_output").modify_bg(
            gtk.STATE_NORMAL, gtk.gdk.color_parse("green")
        )

    def set_blacklist_file(self, file):
        self.blacklist_file = file


    def set_output_file(self, file):
        self.cadbiom_file = file


    def import_BioPAX_data(self, widget):
        """
        Import from BioPAX data on triplestore
        """
        from biopax2cadbiom import biopax_converter

        # Return if output file is not set
        if self.cadbiom_file == '':
            red = gtk.gdk.color_parse("red")
            self.wtree.get_widget("button_output").modify_bg(
                gtk.STATE_NORMAL, red)
            return

        # Get parameters
        # List of graphs URIs
        self.graph_uris = \
            self.wtree.get_widget('text_graphs_uris').get_text().split(',')

        # Triplestore URL
        self.triplestore_url = \
            self.wtree.get_widget('text_triplestore_url').get_text()

        # BioPAX level
        biopax2 = self.wtree.get_widget('radio_biopax_level2').get_active()
        self.graph_uris.append(
            'http://biopax.org/lvl2' if biopax2 else 'http://biopax.org/lvl3'
        )

        # BioPAX convertion
        self.convert_full_graph = \
            self.wtree.get_widget('checkbox_convert_full_graph').get_active()
        self.full_compartments_names = \
            self.wtree.get_widget('checkbox_full_compartment_names').get_active()
        self.remove_strongly_connected_components = \
            self.wtree.get_widget('checkbox_remove_scc').get_active()

        # Color the button in green when the form is valid
        green = gtk.gdk.color_parse("green")
        self.wtree.get_widget("button_make_model").modify_bg(
            gtk.STATE_NORMAL, green)
        self.wtree.get_widget("button_make_model").set_label(
            "Make model. Please wait...")

        # Do the magick
        # Build parameters for biopax2cadbiom
        params = {
            'cadbiomFile': self.cadbiom_file,
            'convertFullGraph': self.convert_full_graph,
            'listOfGraphUri': self.graph_uris,
            'pickleBackup': False,
            'pickleDir': '', # don't care, because pickleBackup = False
            'fullCompartmentsNames': self.full_compartments_names,
            'blacklist': self.blacklist_file,
            'triplestore': self.triplestore_url,
            'no_scc_fix': True, # StartNodes are added here
        }

        try:
            # StartNodes are added here, (not in biopax2cadbiom)
            # so we can load the model in memory
            biopax_converter.main(params)

            if self.remove_strongly_connected_components:
                model = model_corrections.add_start_nodes(self.cadbiom_file)
            else:
                model = MakeModelFromXmlFile(self.cadbiom_file).get_model()

            model.modified = False # ?????
            self.charter.add_edit_mvc(model.name, model)
            self.destroy()

        except Exception as e:

            d_err = DisplayError(
                None, "Biopax2cadbiom",
                "Biopax2cadbiom exception: " + \
                e.__class__.__name__ + ". Please check logs.\n" + \
                traceback.format_exc()
            )
            d_err.main_window.set_modal(True)


    def on_destroy(self, widget):
        """
        standard destroy callback
        """
        if self.main_window:
            self.main_window.destroy()

    def destroy(self):
        """
        for parents
        """
        self.on_destroy(None)

    def on_escape(self, widget, event):
        """On ESC key_press_event, destroy this window."""
        if gtk.gdk.keyval_name(event.keyval) == "Escape":
            self.destroy()



""" 
This module contains the PyGTKCodeBuffer-class. This class is a 
specialisation of the gtk.TextBuffer and enables syntax-highlighting for 
PyGTK's TextView-widget. 
    
To use the syntax-highlighting feature you have load a syntax-definition or
specify your own. To load one please read the docs for the SyntaxLoader()
class. 
"""


# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import gtk
import pango
import re
import sys
import os.path
import xml.sax
import imp
from xml.sax.handler import ContentHandler
from xml.sax.saxutils import unescape

__version__ = "1.0RC2"
__author__  = "Hannes Matuschek <hmatuschek@gmail.com>"


# defined the default styles
DEFAULT_STYLES = {
    'DEFAULT':      {'font': 'monospace'},
    'comment':      {'foreground': '#0000FF'},
    'preprocessor': {'foreground': '#A020F0'},
    'keyword':      {'foreground': '#A52A2A',
                     'weight': pango.WEIGHT_BOLD},
    'special':      {'foreground': 'turquoise'},
    'mark1':        {'foreground': '#008B8B'},
    'mark2':        {'foreground': '#6A5ACD'},
    'string':       {'foreground': '#FF00FF'},
    'number':       {'foreground': '#FF00FF'},
    'datatype':     {'foreground': '#2E8B57',
                     'weight': pango.WEIGHT_BOLD},
    'function':     {'foreground': '#008A8C'},

    'link':         {'foreground': '#0000FF',
                     'underline': pango.UNDERLINE_SINGLE}}


        


def _main_is_frozen():
    """ Internal used function. """
    return (hasattr(sys, "frozen") or # new py2exe
            hasattr(sys, "importers") # old py2exe
            or imp.is_frozen("__main__")) # tools/freeze


if _main_is_frozen():
    this_module_path = os.path.dirname(sys.executable)
else:
    this_module_path = os.path.abspath(os.path.dirname(__file__))


# defines default-search paths for syntax-files 
SYNTAX_PATH = [ os.path.join('.', 'syntax'),
                this_module_path,
                os.path.join(os.path.expanduser('~'),".pygtkcodebuffer"),
                os.path.join(sys.prefix,"share","pygtkcodebuffer","syntax")]
         

# enable/disable debug-messages
DEBUG_FLAG  = False


#
# Some log functions...
#   (internal used)
def _log_debug(msg):
    if not DEBUG_FLAG:
        return
    sys.stderr.write("DEBUG: ")
    sys.stderr.write(msg)
    sys.stderr.write("\n")
    
def _log_warn(msg):
    sys.stderr.write("WARN: ")
    sys.stderr.write(msg)
    sys.stderr.write("\n")
        
def _log_error(msg):
    sys.stderr.write("ERROR: ")
    sys.stderr.write(msg)
    sys.stderr.write("\n")
        



def add_syntax_path(path_or_list):
    """ This function adds one (string) or many (list of strings) paths to the 
        global search-paths for syntax-files. """
    global SYNTAX_PATH
    # handle list of strings
    if isinstance(path_or_list, (list, tuple)):
        for i in range(len(path_or_list)):
            SYNTAX_PATH.insert(0, path_or_list[-i])
    # handle single string
    elif isinstance(path_or_list, basestring):
        SYNTAX_PATH.insert(0, path_or_list)
    # handle attr-error
    else:
        raise TypeError, "Argument must be path-string or list of strings"
        
        
        
class Pattern:
    """ More or less internal used class representing a pattern. You may use 
        this class to "hard-code" your syntax-definition. """

    def __init__(self, regexp, style="DEFAULT", group=0, flags=""):
        """ The constructor takes at least on argument: the regular-expression.
            
            The optional kwarg style defines the style applied to the string
            matched by the regexp. 
            
            The kwarg group may be used to define which group of the regular 
            expression will be used for highlighting (Note: This means that only
            the selected group will be highlighted but the complete pattern must
            match!)
            
            The optional kwarg flags specifies flags for the regular expression.
            Look at the Python lib-ref for a list of flags and there meaning."""
        # assemble re-flag
        flags += "ML"; flag   = 0
        
        _log_debug("init rule %s -> %s (%s)"%(regexp, style, flags))
        
        for char in flags:
            if char == 'M': flag |= re.M
            if char == 'L': flag |= re.L
            if char == 'S': flag |= re.S
            if char == 'I': flag |= re.I
            if char == 'U': flag |= re.U
            if char == 'X': flag |= re.X

        # compile re        
        try: self._regexp = re.compile(regexp, flag)
        except re.error, e: 
            raise Exception("Invalid regexp \"%s\": %s"%(regexp,str(e)))

        self._group  = group
        self.tag_name = style
        
        
    def __call__(self, txt, start, end):
        m = self._regexp.search(txt)
        if not m: return None
        
        mstart, mend = m.start(self._group), m.end(self._group)
        s = start.copy(); s.forward_chars(mstart)
        e = start.copy(); e.forward_chars(mend)
        
        return (s,e)    
    


class KeywordList(Pattern):
    """ This class may be used for hard-code a syntax-definition. It specifies 
        a pattern for a keyword-list. This simplifies the definition of 
        keyword-lists. """

    def __init__(self, keywords, style="keyword", flags=""):
        """ The constructor takes at least on argument: A list of strings 
            specifying the keywords to highlight. 
            
            The optional kwarg style specifies the style used to highlight these
            keywords. 
            
            The optional kwarg flags specifies the flags for the 
            (internal generated) regular-expression. """
        # MLB change for immediate recognition
        #regexp = "(?:\W|^)(%s)\W"%("|".join(keywords),)
        regexp = "(?:\W|^)(%s)(\W|$)"%("|".join(keywords),)
        Pattern.__init__(self, regexp, style, group=1, flags=flags)
        
        
        
class String:
    """ This class may be used to hard-code a syntax-definition. It simplifies 
        the definition of a "string". A "string" is something that consists of
        a start-pattern and an end-pattern. The end-pattern may be content of 
        the string if it is escaped. """

    def __init__(self, starts, ends, escape=None, style="string"):
        """ The constructor needs at least two arguments: The start- and 
            end-pattern. 
            
            The optional kwarg escape specifies a escape-sequence escaping the 
            end-pattern.
            
            The optional kwarg style specifies the style used to highlight the
            string. """
        try:
            self._starts  = re.compile(starts)
        except re.error, e: 
            raise Exception("Invalid regexp \"%s\": %s"%(regexp,str(e)))
        
        if escape:
            end_exp = "[^%(esc)s](?:%(esc)s%(esc)s)*%(end)s"
            end_exp = end_exp%{'esc':escape*2,'end':ends}
        else:
            end_exp = ends

        try:     
            self._ends    = re.compile(end_exp)
        except re.error, e: 
            raise Exception("Invalid regexp \"%s\": %s"%(regexp,str(e)))

        self.tag_name = style


    def __call__(self, txt, start, end):
        start_match = self._starts.search(txt)
        if not start_match: return
        
        start_it = start.copy()
        start_it.forward_chars(start_match.start(0))
        end_it   = end.copy()
        
        end_match = self._ends.search(txt, start_match.end(0)-1)
        if end_match:
            end_it.set_offset(start.get_offset()+end_match.end(0))            
            
        return  start_it, end_it


        
class LanguageDefinition:
    """ This class is a container class for all rules (Pattern, KeywordList, 
        ...) specifying the language. You have to used this class if you like
        to hard-code your syntax-definition. """
        
    def __init__(self, rules):
        """ The constructor takes only one argument: A list of rules (i.e 
            Pattern, KeywordList and String). """
        self._grammar = rules
        self._styles = dict()
                
                
    def __call__(self, buf, start, end=None):
        # if no end given -> end of buffer
        if not end: end = buf.get_end_iter()
    
        mstart = mend = end
        mtag   = None
        txt = buf.get_slice(start, end)    
        
        # search min match
        for rule in self._grammar:
            # search pattern
            m = rule(txt, start, end)            
            if not m: continue
            
            # prefer match with smallest start-iter 
            if m[0].compare(mstart) < 0:
                mstart, mend = m
                mtag = rule.tag_name
                continue
            
            if m[0].compare(mstart)==0 and m[1].compare(mend)>0:
                mstart, mend = m
                mtag = rule.tag_name
                continue

        return (mstart, mend, mtag)                


    def get_styles(self):
        return self._styles
        



class SyntaxLoader(ContentHandler, LanguageDefinition):
    """ This class loads a syntax definition. There have to be a file
        named LANGUAGENAME.xml in one of the directories specified in the
        global path-list. You may add a directory using the add_syntax_path()
        function. """
    
    # some translation-tables for the style-defs:
    style_weight_table =    {'ultralight': pango.WEIGHT_ULTRALIGHT,
                             'light': pango.WEIGHT_LIGHT,
                             'normal': pango.WEIGHT_NORMAL,
                             'bold':   pango.WEIGHT_BOLD,
                             'ultrabold': pango.WEIGHT_ULTRABOLD,
                             'heavy': pango.WEIGHT_HEAVY}
    style_variant_table =   {'normal': pango.VARIANT_NORMAL,
                             'smallcaps': pango.VARIANT_SMALL_CAPS}
    style_underline_table = {'none': pango.UNDERLINE_NONE,
                             'single': pango.UNDERLINE_SINGLE,
                             'double': pango.UNDERLINE_DOUBLE}
    style_style_table =     {'normal': pango.STYLE_NORMAL,
                             'oblique': pango.STYLE_OBLIQUE,
                             'italic': pango.STYLE_ITALIC}                                                   
    style_scale_table =     {
                            'xx_small': pango.SCALE_XX_SMALL,
                            'x_small':  pango.SCALE_X_SMALL,
                            'small':  pango.SCALE_SMALL,
                            'medium':  pango.SCALE_MEDIUM,
                            'large':  pango.SCALE_LARGE,
                            'x_large':  pango.SCALE_X_LARGE,
                            'xx_large': pango.SCALE_XX_LARGE,
                            }
                          
                          
    def __init__(self, lang_name):
        """ The constructor takes only one argument: the language name.
            The constructor tries to load the syntax-definition from a 
            syntax-file in one directory of the global path-list. 
            
            An instance of this class IS a LanguageDefinition. You can pass it
            to the constructor of the CodeBuffer class. """
        LanguageDefinition.__init__(self, [])
        ContentHandler.__init__(self)

        # search for syntax-files:
        fname = None
        for syntax_dir in SYNTAX_PATH:
            fname = os.path.join(syntax_dir, "%s.xml"%lang_name)
            if os.path.isfile(fname): break

        _log_debug("Loading syntaxfile %s"%fname)
        
        if not os.path.isfile(fname):
            raise Exception("No snytax-file for %s found!"%lang_name)
            
        xml.sax.parse(fname, self)
        
        
    # Dispatch start/end - document/element and chars        
    def startDocument(self):
        self.__stack   = []
                
    def endDocument(self):
        del self.__stack
        
    def startElement(self, name, attr):
        self.__stack.append( (name, attr) )
        if hasattr(self, "start_%s"%name):
            handler = getattr(self, "start_%s"%name)
            handler(attr)
    
    def endElement(self, name):
        if hasattr(self, "end_%s"%name):
            handler = getattr(self, "end_%s"%name)
            handler()
        del self.__stack[-1]

    def characters(self, txt):
        if not self.__stack: return
        name, attr = self.__stack[-1]
        
        if hasattr(self, "chars_%s"%name):
            handler = getattr(self, "chars_%s"%name)
            handler(txt)
            

    # Handle regexp-patterns
    def start_pattern(self, attr):
        self.__pattern = ""
        self.__group   = 0
        self.__flags   = ''
        self.__style   = attr['style']
        if 'group' in attr.keys(): self.__group = int(attr['group'])
        if 'flags' in attr.keys(): self.__flags = attr['flags']
        
    def end_pattern(self):
        rule = Pattern(self.__pattern, self.__style, self.__group, self.__flags)
        self._grammar.append(rule)
        del self.__pattern
        del self.__group
        del self.__flags
        del self.__style
        
    def chars_pattern(self, txt):
        self.__pattern += unescape(txt)
                    

    # handle keyword-lists
    def start_keywordlist(self, attr):
        self.__style = "keyword"
        self.__flags = ""
        if 'style' in attr.keys():
            self.__style = attr['style']
        if 'flags' in attr.keys():
            self.__flags = attr['flags']
        self.__keywords = []
        
    def end_keywordlist(self):
        kwlist = KeywordList(self.__keywords, self.__style, self.__flags)
        self._grammar.append(kwlist)
        del self.__keywords
        del self.__style
        del self.__flags
        
    def start_keyword(self, attr):
        self.__keywords.append("")
    
    def end_keyword(self):
        if not self.__keywords[-1]:
            del self.__keywords[-1]
                
    def chars_keyword(self, txt):
        parent,pattr = self.__stack[-2]
        if not parent == "keywordlist": return
        self.__keywords[-1] += unescape(txt)


    #handle String-definitions
    def start_string(self, attr):
        self.__style = "string"
        self.__escape = None
        if 'escape' in attr.keys():
            self.__escape = attr['escape']
        if 'style' in attr.keys():
            self.__style = attr['style']
        self.__start_pattern = ""
        self.__end_pattern = ""

    def end_string(self):
        strdef = String(self.__start_pattern, self.__end_pattern,
                        self.__escape, self.__style)
        self._grammar.append(strdef)
        del self.__style
        del self.__escape
        del self.__start_pattern
        del self.__end_pattern
        
    def chars_starts(self, txt):
        self.__start_pattern += unescape(txt)
        
    def chars_ends(self, txt):
        self.__end_pattern += unescape(txt)


    # handle style
    def start_style(self, attr):
        self.__style_props = dict()
        self.__style_name = attr['name']
        
    def end_style(self):
        self._styles[self.__style_name] = self.__style_props
        del self.__style_props
        del self.__style_name
        
    def start_property(self, attr):
        self.__style_prop_name = attr['name']
        
    def chars_property(self, value):
        value.strip()
        
        # convert value
        if self.__style_prop_name in ['font','foreground','background',]:
            pass
            
        elif self.__style_prop_name == 'variant':
            if not value in self.style_variant_table.keys():
                Exception("Unknown style-variant: %s"%value)
            value = self.style_variant_table[value]
                
        elif self.__style_prop_name == 'underline':
            if not value in self.style_underline_table.keys():
                Exception("Unknown underline-style: %s"%value)
            value = self.style_underline_table[value]
                
        elif self.__style_prop_name == 'scale':
            if not value in self.style_scale_table.keys():
                Exception("Unknown scale-style: %s"%value)
            value = self.style_scale_table[value]

        elif self.__style_prop_name == 'weight':
            if not value in self.style_weight_table.keys():
                Exception("Unknown style-weight: %s"%value)
            value = self.style_weight_table[value]
                
        elif self.__style_prop_name == 'style':
            if not value in self.style_style_table[value]:
                Exception("Unknwon text-style: %s"%value)
            value = self.style_style_table[value]
                
        else:
            raise Exception("Unknown style-property %s"%self.__style_prop_name)

        # store value            
        self.__style_props[self.__style_prop_name] = value
            
            
            
            
class CodeBuffer(gtk.TextBuffer):
    """ This class extends the gtk.TextBuffer to support syntax-highlighting. 
        You can use this class like a normal TextBuffer. """
        
    def __init__(self, table=None, lang=None, styles={}):
        """ The constructor takes 3 optional arguments. 
        
            table specifies a tag-table associated with the TextBuffer-instance.
            This argument will be passed directly to the constructor of the 
            TextBuffer-class. 
            
            lang specifies the language-definition. You have to load one using
            the SyntaxLoader-class or you may hard-code your syntax-definition 
            using the LanguageDefinition-class. 
            
            styles is a dictionary used to extend or overwrite the default styles
            provided by this module (DEFAULT_STYLE) and any language specific 
            styles defined by the LanguageDefinition. """
        gtk.TextBuffer.__init__(self, table)

        # default styles    
        self.styles = DEFAULT_STYLES
                       
        # update styles with lang-spec:
        if lang:
            self.styles.update(lang.get_styles())
            
        # update styles with user-defined
        self.styles.update(styles)
        
        # create tags
        for name, props in self.styles.items():
            style = dict(self.styles['DEFAULT'])    # take default
            style.update(props)                     # and update with props
            self.create_tag(name, **style)
        
        # store lang-definition
        self._lang_def = lang
        
        self.connect_after("insert-text", self._on_insert_text)
        self.connect_after("delete-range", self._on_delete_range)
        self.connect('apply-tag', self._on_apply_tag)
        
        self._apply_tags = False
                
                
    def _on_apply_tag(self, buf, tag, start, end):
        # FIXME This is a hack! It allows apply-tag only while
        #       _on_insert_text() and _on_delete_range()
        if not self._apply_tags:
            self.emit_stop_by_name('apply-tag')
            return True
            
        _log_debug("tag \"%s\" as %s"%(self.get_slice(start,end), tag.get_property("name")))
            
                            
    def _on_insert_text(self, buf, it, text, length):
        # if no syntax defined -> nop
        if not self._lang_def: return False
        
        it = it.copy()
        it.backward_chars(length)
        
        if not it.begins_tag():
            it.backward_to_tag_toggle(None)
            _log_debug("Not tag-start -> moved iter to %i (%s)"%(it.get_offset(), it.get_char()))

        if it.begins_tag(self.get_tag_table().lookup("DEFAULT")):
            it.backward_to_tag_toggle(None)
            _log_debug("Iter at DEFAULT-start -> moved to %i (%s)"%(it.get_offset(), it.get_char()))
            
        self._apply_tags = True    
        self.update_syntax(it)        
        self._apply_tags = False
        
        
    def _on_delete_range(self, buf, start, end):
        # if no syntax defined -> nop
        if not self._lang_def: return False

        start = start.copy()
        if not start.begins_tag():
            start.backward_to_tag_toggle(None)
    
        self._apply_tags = True                
        self.update_syntax(start)        
        self._apply_tags = False
        
    
    def update_syntax(self, start, end=None):
        """ More or less internal used method to update the 
            syntax-highlighting. """
        # if no lang set    
        if not self._lang_def: return             
        _log_debug("Update syntax from %i"%start.get_offset())
            
        # if not end defined
        if not end: end = self.get_end_iter()
        
        # We do not use recursion -> long files exceed rec-limit!
        finished = False
        while not finished: 
            # search first rule matching txt[start..end]            
            mstart, mend, tagname = self._lang_def(self, start, end)
            # optimisation: if mstart-mend is allready tagged with tagname 
            #   -> finished
            if tagname:     #if something found
                tag = self.get_tag_table().lookup(tagname)
                if mstart.begins_tag(tag) and mend.ends_tag(tag) and not mstart.equal(start):
                    self.remove_all_tags(start,mstart)
                    self.apply_tag_by_name("DEFAULT", start, mstart)
                    _log_debug("Optimized: Found old tag at %i (%s)"%(mstart.get_offset(), mstart.get_char()))
                    # finish
                    finished = True
                    continue
                    
            # remove all tags from start..mend (mend == buffer-end if no match)        
            self.remove_all_tags(start, mend)
            # make start..mstart = DEFAUL (mstart == buffer-end if no match)
            if not start.equal(mstart):
                _log_debug("Apply DEFAULT")
                self.apply_tag_by_name("DEFAULT", start, mstart)                
        
            # nothing found -> finished
            if not tagname: 
                finished = True
                continue
        
            # apply tag
            _log_debug("Apply %s"%tagname)
            self.apply_tag_by_name(tagname, mstart, mend)

            start = mend
            
            if start == end: 
                finished = True                 
                continue
                
        
    def reset_language(self, lang_def):
        """ Reset the currently used language-definition. """
        # remove all tags from complete text
        start = self.get_start_iter()
        self.remove_all_tags(start, self.get_end_iter())
        # store lexer
        self._lang_def = lang_def
        # update styles from lang_def:
        if self._lang_def:
            self.update_styles(self._lang_def.get_styles())
        # and ...
        self._apply_tags = True
        self.update_syntax(start)
        self._apply_tags = False
        
        
    def update_styles(self, styles):
        """ Update styles. This method may be used to reset any styles at
            runtime. """
        self.styles.update(styles)
        
        table = self.get_tag_table()
        for name, props in styles.items():
            style = self.styles['DEFAULT']
            style.update(props)
            # if tagname is unknown:
            if not table.lookup(name):
                _log_debug("Create tag: %s (%s)"%(name, style))
                self.create_tag(name, **style) 
            else: # update tag
                tag = table.lookup(name)
                _log_debug("Update tag %s with (%s)"%(name, style))
                map(lambda i: tag.set_property(i[0],i[1]), style.items())

                        



## Filename    : charter.py
## Author(s)   : Geoffroy Andrieux
## Created     : 03/2010
## Revision    :
## Source      :
##
## Copyright 2010 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA/IRSET
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Michel Le Borgne, Nolwenn Le Meur
##
"""
Main class for Cadbiom gui
"""

import gtk
import webbrowser

from gtk.gdk import screen_height, screen_width, \
                  COLORSPACE_RGB, colormap_get_system, Pixbuf
from cadbiom.antlr3 import ANTLRFileStream, CommonTokenStream

from charter_info import CharterInfo
from edit_mvc import EditMVC
from utils.warn import  confirm, DialogEntry, cancel_warn
from utils.notebookUtils import create_custom_tab
from utils.text_page import BioSignalEditor
from utils.reporter import CompilReporter
from utils.fileHandling import FileChooser

from chart_controler import ChartClipboard
from chart_simulator.chart_simul_controler import \
                     ChartSimulControler, DisplayError
from chart_checker.chart_checker_controler  import ChartChecker
from chart_misc_widgets import SearchManager, SearchFrontier, \
                      LegendWindow, ImportPIDParam, ImportBioPAXParams
from cadbiom_gui.gt_gui.chart_static.chart_stat_controler import \
    STATWindow, SCCWindow, BAGWindow

from cadbiom.models.guard_transitions.chart_model import ChartModel
from cadbiom.models.guard_transitions.translators.chart_lang import LangVisitor
from cadbiom.models.guard_transitions.translators.chart_xml import \
                  XmlVisitor, MakeModelFromXmlFile, \
                  MakeModelFromXmlString, XmlException
from cadbiom.models.guard_transitions.translators.cadlangLexer import cadlangLexer
from cadbiom.models.guard_transitions.translators.cadlangParser import cadlangParser

# Custom imports
from cadbiom.models.guard_transitions.analyser.static_analysis \
    import StaticAnalyzer

from layout import LayoutVisitor

import pkg_resources

class Charter(object):
    """
    charter main class
    """

    def __init__(self, cad):
        self.ident = "charter"
        self.cad_manager = cad

        # Gestion of the destroy of all subwindows opened from this one
        self.subwindows = set()

        # simul options
        self.simul_strict = True
        self.sim_flat_graph = False

        # common clipboard
        self.clipboard = ChartClipboard()
        # gui
        # Set the Glade file
        template = pkg_resources.resource_filename(
            __name__,
            "chart_glade/charter.glade"
        )
        self.wtree = gtk.glade.XML(template)
        # Get the Main Window, and connect the "destroy" event
        self.main_window = self.wtree.get_widget("TopModel")
        if self.main_window:
            self.main_window.connect("delete_event", self.on_destroy)
        h_scr = screen_height()
        w_scr = screen_width()
        self.h_scr = int(h_scr)
        self.w_scr = int(w_scr)
        self.main_window.resize(int(w_scr * 0.9), int(h_scr * 0.9))
        self.constraint_window = None

        # Accelerators
        self.my_accelerators = gtk.AccelGroup()
        self.main_window.add_accel_group(self.my_accelerators)

        # menu bar
        menu_bar = self.wtree.get_widget("menubar1")
        menu_bar.set_size_request(0, 25)

        menu_item =  self.wtree.get_widget("doc_menu")
        menu_item.connect("activate", self.show_doc)
        menu_item =  self.wtree.get_widget("legend_menu")
        menu_item.connect("activate", self.show_legend)

        menu_item =  self.wtree.get_widget("New_model")
        self.add_accelerator(menu_item, "<Control>n")
        menu_item.connect("activate", self.new_charts)
        menu_item =  self.wtree.get_widget("Import_cadbiom")
        self.add_accelerator(menu_item, "<Control>o")
        menu_item.connect("activate", self.choose_xml_file)
        menu_item =  self.wtree.get_widget("Import_BioPAX")
        self.add_accelerator(menu_item, "<Control>b")
        menu_item.connect("activate", self.choose_BioPAX_file)
        menu_item =  self.wtree.get_widget("Import_PID")
        menu_item.connect("activate", self.choose_pid_file)
        menu_item =  self.wtree.get_widget("Import_CadLang")
        menu_item.connect("activate", self.choose_cadlang_file)
        menu_item =  self.wtree.get_widget("Export_cadbiom")
        self.add_accelerator(menu_item, "<Control>s")
        menu_item.connect("activate", self.export_to_xml)
        menu_item =  self.wtree.get_widget("Export_cadlang")
        menu_item.connect("activate", self.export_to_lang)
        menu_item =  self.wtree.get_widget("Export_picture")
        menu_item.connect("activate", self.export_picture)
        # Add Quit button & shortcuts
        menu =  self.wtree.get_widget("menu2")
        menu.append(gtk.SeparatorMenuItem())
        menu_item = gtk.ImageMenuItem(gtk.STOCK_QUIT)
        self.add_accelerator(menu_item, "<Control>q")
        menu_item.connect("activate", self.on_destroy, None)
        menu.append(menu_item)

        menu_item =  self.wtree.get_widget("Hierarchical_TB")
        menu_item.connect("activate", self.do_layout, "hierarchical_TB")
        menu_item =  self.wtree.get_widget("Hierarchical_LR")
        menu_item.connect("activate", self.do_layout, "hierarchical_LR")
        menu_item =  self.wtree.get_widget("neato")
        menu_item.connect("activate", self.do_layout, "neato")
        menu_item =  self.wtree.get_widget("fdp")
        menu_item.connect("activate", self.do_layout, "fdp")
        menu_item =  self.wtree.get_widget("twopi")
        menu_item.connect("activate", self.do_layout, "twopi")
        menu_item =  self.wtree.get_widget("circo")
        menu_item.connect("activate", self.do_layout, "circo")

        menu_item =  self.wtree.get_widget("model_infos")
        menu_item.connect("activate", self.on_stats_info)
        menu_item =  self.wtree.get_widget("scc_search")
        menu_item.connect("activate", self.on_frontier_scc)
        menu_item =  self.wtree.get_widget("basal_activated_genes")
        menu_item.connect("activate", self.on_basal_activated_genes)
        menu_item =  self.wtree.get_widget("dependance_graph")

        menu_item =  self.wtree.get_widget("transitionGraph")
        menu_item.connect("activate", self.on_dependency_graph)
        menu_item =  self.wtree.get_widget("dependencyGraph")
        menu_item.connect("activate", self.on_dependency_graph)
        menu_item =  self.wtree.get_widget("fullDependencyGraph")
        menu_item.connect("activate", self.on_dependency_graph)

        # Model Handling buttons
        button = self.wtree.get_widget("Check_button")
        button.set_label(button.get_label() + " (F8)")
        self.add_accelerator(button, "F8")
        button.connect("clicked", self.on_check)

        button = self.wtree.get_widget("Simu_button")
        button.set_label(button.get_label() + " (F9)")
        self.add_accelerator(button, "F9")
        button.connect("clicked", self.on_simulate)

        # graph drawing controler buttons (not connected)
        self.graph_notebook = self.wtree.get_widget("graph_notebook")
        self.graph_notebook.connect("switch-page",
                                    self.switch_graph_page_callback)
        self.create_drawing_buttons()

        # edit MVC
        self.ei_id = 0
        self.edit_mvc_list = []
        self.current_edit_mvc = None

        # chart info
        chart_info_frame = self.wtree.get_widget("informations_frame")
        self.chart_info = CharterInfo(chart_info_frame)
        self.chart_info.trans_info.show.connect("clicked", self.display_states)
        mtk_notebook = self.wtree.get_widget("MTK_notebook")
        # search
        self.search_area = SearchManager(self, mtk_notebook, "Simple node list")
        self.search_frontier = SearchFrontier(self, mtk_notebook, "Frontier")

        # overview
        self.overview_window = self.wtree.get_widget("Overview")


    def check_curent_model(function):
        """Decorator that checks if there is a current model.

        It is used before the call of functions that do some processing on
        models.

        If there is no model, the function juste returns None.
        """

        def modified_func(self, *args, **kwargs):
            """Returned modified function"""

            if not self.current_edit_mvc:
                return

            return function(self, *args, **kwargs)
        return modified_func

    def refresh(self):
        """
        TODO
        """
        pass

    def hide(self):
        """
        Useful ?
        """
        self.main_window.hide_all()

    def create_drawing_buttons(self):
        """
        create the buttons to control drawing
        """
        self.button_handlers = dict()
        c_box = self.wtree.get_widget("controler_box")

        but = gtk.Button(label="Select")
        c_box.add(but)
        self.button_handlers["Select"] = (but, -1)

        but = gtk.Button(label="InputNode (F1)")
        self.add_accelerator(but, "F1")
        self.add_image(but, "input_node")
        c_box.add(but)
        self.button_handlers["InputNode"] = (but, -1)

        but = gtk.Button(label="SimpleNode (F2)")
        self.add_accelerator(but, "F2")
        self.add_image(but, "simple_node")
        c_box.add(but)
        self.button_handlers["SimpleNode"] = (but, -1)

        but = gtk.Button(label="PermNode (F3)")
        self.add_accelerator(but, "F3")
        self.add_image(but, "permanent_node")
        c_box.add(but)
        self.button_handlers["PermNode"] = (but, -1)

        but = gtk.Button(label="StartNode (F4)")
        self.add_accelerator(but, "F4")
        self.add_image(but, "start_node")
        c_box.add(but)
        self.button_handlers["StartNode"] = (but, -1)

        but = gtk.Button(label="TrapNode (F5)")
        self.add_accelerator(but, "F5")
        self.add_image(but, "trap_node")
        c_box.add(but)
        self.button_handlers["TrapNode"] = (but, -1)

        but = gtk.Button(label="Transition (F6)")
        self.add_accelerator(but, "F6")
        self.add_image(but, "transition")
        c_box.add(but)
        self.button_handlers["Transition"] = (but, -1)

        but = gtk.Button(label="View constraints (F7)")
        self.add_accelerator(but, "F7")
        c_box.add(but)
        self.button_handlers["Constraints"] = (but, -1)

        # zoom
        but = self.wtree.get_widget("zoomp")
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_ZOOM_IN, 3)
        but.set_image(image)
        self.button_handlers["zoom_p"] = (but, -1)
        but = self.wtree.get_widget("zoomm")
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_ZOOM_OUT, 3)
        but.set_image(image)
        self.button_handlers["zoom_m"] = (but, -1)


    def drawing_button_connect(self):
        """
        Connect the buttons to the controler of the graphic view
        Remember for each button the handler
        """
        (but, bhan) = self.button_handlers["Select"]
        bhan = but.connect("clicked",
                         self.current_edit_mvc.controler.set_action_select)
        self.button_handlers["Select"] = (but, bhan)

        (but, bhan) = self.button_handlers["InputNode"]
        bhan = but.connect("clicked",
                    self.current_edit_mvc.controler.set_action_new_input_node)
        self.button_handlers["InputNode"] = (but, bhan)

        (but, bhan) = self.button_handlers["SimpleNode"]
        bhan = but.connect("clicked",
                    self.current_edit_mvc.controler.set_action_new_simple_node)
        self.button_handlers["SimpleNode"] = (but, bhan)

        (but, bhan) = self.button_handlers["PermNode"]
        bhan = but.connect("clicked",
                    self.current_edit_mvc.controler.set_action_new_perm_node)
        self.button_handlers["PermNode"] = (but, bhan)


        (but, bhan) = self.button_handlers["StartNode"]
        bhan = but.connect("clicked",
                    self.current_edit_mvc.controler.set_action_new_start_node)
        self.button_handlers["StartNode"] = (but, bhan)

        (but, bhan) = self.button_handlers["TrapNode"]
        bhan = but.connect("clicked",
                    self.current_edit_mvc.controler.set_action_new_trap_node)
        self.button_handlers["TrapNode"] = (but, bhan)

        (but, bhan) = self.button_handlers["Transition"]
        bhan = but.connect("clicked",
                    self.current_edit_mvc.controler.set_action_new_transition)
        self.button_handlers["Transition"] = (but, bhan)

        (but, bhan) = self.button_handlers["Constraints"]
        bhan = but.connect("clicked", self.edit_constraints)
        self.button_handlers["Constraints"] = (but, bhan)

        (but, bhan) = self.button_handlers["zoom_p"]
        bhan = but.connect("clicked", self.current_edit_mvc.zoom_plus)
        self.button_handlers["zoom_p"] = (but, bhan)

        (but, bhan) = self.button_handlers["zoom_m"]
        bhan = but.connect("clicked", self.current_edit_mvc.zoom_minus)
        self.button_handlers["zoom_m"] = (but, bhan)


    def add_image(self, widget, image_file):
        """Add image to a Button
        from http://www.pygtk.org/docs/pygtk/gtk-stock-items.html
        """
        image = gtk.Image()
        # Get resource
        template = pkg_resources.resource_filename(
            __name__,
            'images/' + image_file + '.png',
        )
        image.set_from_file(template)
        widget.set_image(image)

    def add_accelerator(self, widget, accelerator, signal="activate"):
        """Adds a keyboard shortcut"""
        if accelerator is not None:
            #if DEBUG:
                #print accelerator, widget.get_tooltip_text()
            key, mod = gtk.accelerator_parse(accelerator)
            # VISIBLE: if set, the accelerator is visible in a label
            widget.add_accelerator(signal,
                                   self.my_accelerators,
                                   key, mod,
                                   gtk.ACCEL_VISIBLE,
            )
            # print "The accelerator is well added with the signal " + signal

    def drawing_button_disconnect(self):
        """
        Disconnect graphic editor buttons
        Carefull: do not change value of handler for efficiency reason
        """
        for k in self.button_handlers.keys():
            bhan = self.button_handlers[k]
            bhan[0].disconnect(bhan[1])

    def drawing_button_disable(self):
        """
        Disconnect graphic editor buttons
        Carefull: do not change value of handler for efficiency reason
        """
        for k in self.button_handlers.keys():
            bhan = self.button_handlers[k]
            bhan[0].set_sensitive(False)

    def drawing_button_unable(self):
        """
        Disconnect graphic editor buttons
        Carefull: do not change value of handler for efficiency reason
        """
        for k in self.button_handlers.keys():
            bhan = self.button_handlers[k]
            bhan[0].set_sensitive(True)

    def on_destroy(self, widget, _dummy_event):
        """
        destroy if everything OK
        """
        # check if some models are modified
        for emvc in self.edit_mvc_list:
            if emvc.model.is_modified():
                name = emvc.model.name
                ask = False
                ask = confirm(None, 'Model '+name+' is modified - Quit anyway?')
                if not ask:
                    return True
            emvc.clean_subwin()
        gtk.main_quit()

    def show(self):
        """
        TODO check if useful
        """
        self.main_window.show_all()


    def delete_tab_callback(self, widget):
        """
        Delete current edit_MVC and corresponding view
        """
        self.remove_edit_mvc()



    def add_edit_mvc(self, model_name, model=None, layout=False):
        """
        create, register and display a new edit_mvc - close all windows
        associated with a previous edit_mvc
        """

        # edit MVC
        edm = EditMVC(self, model_name, model, layout)
        (eventbox, button) = create_custom_tab(edm.title)
        button.connect("clicked", self.delete_tab_callback)
        edm.tab_button = button

        # insert in emvc management
        page_index = self.graph_notebook.append_page(edm.viewpage, eventbox)
        self.edit_mvc_list.append(edm)
        self.set_current_edit_mvc(edm)
        self.graph_notebook.set_current_page(page_index)

    def add_display_mvc(self, model_name, model=None, layout=False):
        """
        create, register and display a new edit_mvc
        """
        # edit MVC
        edm = EditMVC(self, model_name, model, layout)
        (eventbox, button) = create_custom_tab(edm.title)
        button.connect("clicked", self.delete_tab_callback)
        edm.tab_button = button

        # insert in emvc management
        page_index = self.graph_notebook.append_page(edm.viewpage, eventbox)
        self.edit_mvc_list.append(edm)
        self.set_current_edit_mvc(edm, False) # do not destroy aux windows
        self.graph_notebook.set_current_page(page_index)

    @check_curent_model
    def remove_edit_mvc(self):
        """
        Remove current edit_MVC
        """
        eimvc = self.current_edit_mvc
        # modified model??
        ask = True
        if not eimvc.model.is_submodel():
            if eimvc.model.is_modified():
                ask = confirm(None, "Modified model. Do you want to delete?")
        if ask:
            # remove from notebook and navigator
            self.search_area.clear()
            self.search_frontier.clear()
            self.overview_window.get_child().get_child().clear()
            # following implies a switch page (thus disconnection)
            self.graph_notebook.remove(eimvc.viewpage)
            # remove from list
            self.edit_mvc_list.remove(eimvc)
            eimvc.clean_subwin()

    def open_macro(self, mnode):
        """
        open a macro state in an other tab
        """
        edm = EditMVC(self, mnode.name)
        edm.model.make_submodel(mnode)
        edm.model.get_root()
        # attach submodel view to whole model
        mnode.model.attach(edm.view)

        (eventbox, button) = create_custom_tab(edm.title)
        button.connect("clicked", self.delete_tab_callback)
        edm.tab_button = button
        page_index = self.graph_notebook.append_page(edm.viewpage, eventbox)
        self.edit_mvc_list.append(edm)
        self.set_current_edit_mvc(edm)
        edm.view.show()
        self.graph_notebook.set_current_page(page_index)

    def set_current_edit_mvc(self, edm, w_destroy=True):
        """
        display a new current edit item
        """
        # disconnect the currrent one if any
        if self.current_edit_mvc:
            self.save_constraints(None)
            self.current_edit_mvc.disconnect(self, w_destroy)

        if edm:
            self.current_edit_mvc = edm
            edm.connect(self)
            # refresh search
            self.search_area.set_model(edm.model)
            self.search_frontier.set_model(edm.model)

    def get_emvc_with_view(self, view):
        """
        Retreive the edit_MVC from the view
        """
        for emvc in self.edit_mvc_list:
            if emvc.view == view:
                return emvc
        return None

    def switch_graph_page_callback(self, widget, page, page_index):
        """
        Standard notebook call back
        """
        if page_index >= 0:
            the_view = self.graph_notebook.get_nth_page(page_index).draw
            edm = self.get_emvc_with_view(the_view)
            if edm: # for the first case
                self.set_current_edit_mvc(edm)
        else: # no more page (in case of delete)
            pass

    def ok_new_text(self, but, dialog):
        """
        Prompt if new model - to give the name
        """
        model_name = dialog.entry.get_text()
        dialog.destroy()
        self.add_edit_mvc(model_name)

    def cancel(self, but, dee):
        """
        what?
        """
        dee.destroy()


    def new_charts(self, widget):
        """
        create a new charts when you click on new
        """

        def on_enter(widget, event):
            """On ESC/Return key_press_event, destroy/valid this window."""
            keyval_name = gtk.gdk.keyval_name(event.keyval)
            if keyval_name == 'Return':
                self.ok_new_text(None, die)

            if keyval_name == 'Escape':
                die.destroy()

        # open an independant gtk.Entry for asking for the name
        die = DialogEntry("Insert a model name")
        die.entry.set_text("default_name")
        die.entry.connect('key_press_event', on_enter)
        die.okb.connect("clicked", self.ok_new_text, die)
        die.cancel.connect("clicked", self.cancel, die)
        die.run()

    def save_choice(self):
        """
        choice of file for save
        """
        dialog = gtk.Dialog("",
                            None,
                            gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                            ("Yes", 0, "No", 1, "Cancel", 2))
        label = gtk.Label("Do yous want to delete previous xml model?")
        dialog.vbox.pack_start(label, True, False, 0)
        label.show()
        response = dialog.run()
        dialog.destroy()
        return response

    @check_curent_model
    def export_to_xml(self, widget):
        """
        open a window to save as xml file
        """
        fch = FileChooser("Export to xml Cadbiom model", "bcx files", "*.bcx")
        fch.set_current_name(self.current_edit_mvc.model.name + ".bcx")
        fch.do_action(self.create_xml_file)


    def create_xml_file(self, xml_file):
        """
        make a xml file with the current model
        """
        if self.current_edit_mvc.model.is_submodel():
            model = self.current_edit_mvc.controler.current_node.model
        else :
            model = self.current_edit_mvc.model
        try:
            xml = XmlVisitor(model)
            mfile = open(xml_file, 'w')
            mfile.write(xml.return_xml())
            mfile.close()
        except XmlException as xec:
            cancel_warn(xec.message)
            return

    def choose_xml_file(self, widget):
        """
        open a window to search xml file
        """
        fch = FileChooser("Import xml Cadbiom model", "bcx files", "*.bcx",
                          save_window=False)
        fch.do_action(self.import_from_xml_file)

    def import_from_xml_file(self, ffile):
        """
        As it says
        """
        self.import_from_xml(ffile, ffile=True)

    def import_from_xml(self, xml_cont, ffile=False):
        """
        open and parse an xml file or from db
        """
        if ffile:
            parsing = MakeModelFromXmlFile(xml_cont)
        else:
            parsing = MakeModelFromXmlString(xml_cont)
        # chart model
        model = parsing.get_model()
        model.modified = False
        self.add_edit_mvc(model.name, model)


    def choose_cadlang_file(self, widget):
        """
        open a window to search xml file
        """
        fch = FileChooser("Import from Cadbiom-chart lang", "cadbiom-l", "*.cal",
                          save_window=False)
        fch.do_action(self.import_from_cl_file)

    def import_from_cl_file(self, ffile):
        """
        As it says
        """
        crep = CompilReporter()
        fstream = ANTLRFileStream(ffile)
        lexer = cadlangLexer(fstream)
        lexer.set_error_reporter(crep)
        parser = cadlangParser(CommonTokenStream(lexer))
        parser.set_error_reporter(crep)
        model = ChartModel(ffile)
        parser.cad_model(model)
        if crep.error:
            DisplayError(crep, ffile)
            return
        model = parser.model
        self.add_edit_mvc(model.name, model, False)
        self.do_layout(None, "hierarchical_LR")


    def choose_pid_file(self, widget):
        """
        open a window to search xml file coming from PID database
        """
        # Pass the parent window to set modal mode on the child
        ImportPIDParam(self, self.main_window)

    def choose_BioPAX_file(self, widget):
        """
        open a window to import BioPAX data from a triplestore
        """
        # Pass the parent window to set modal mode on the child
        ImportBioPAXParams(self, self.main_window)

    @check_curent_model
    def export_to_lang(self, widget):
        """
        Export to CadLang
        """
        fch = FileChooser("Export to Cadbiom-chart lang", "cadbiom-l", "*.cal")
        fch.set_current_name(self.current_edit_mvc.model.name + ".cal")
        fch.do_action(self.compile_to_lang)

    def compile_to_lang(self, file_name):
        """
        Compile in view of exporting
        """
        out = open(file_name, "w")
        # decompiler visitor
        lvi = LangVisitor(out)
        self.current_edit_mvc.model.accept(lvi)
        out.close()

    def show_doc(self, widget):
        """
        Doc
        """
        url = 'http://cadbiom.genouest.org/cw_support.html'
        webbrowser.open(url)
        return

    def show_legend(self, widget):
        """
        Doc
        """
        self.legend = LegendWindow(self)

    @check_curent_model
    def export_picture(self, widget):
        """
        Export a picture (png,jpg) of the current model
        """
        fch = FileChooser("Export picture", "pictures", "*.png")
        fch.set_current_name(self.current_edit_mvc.model.name + ".png")
        fch.get_filter().add_pattern("*.png")
        fch.do_action(self.export_picture_to_file)

    @check_curent_model
    def export_picture_to_file(self, file_name):
        """
        As it says
        """

        edm = self.current_edit_mvc
        fns = file_name.split('.')
        if len(fns)!= 2:
            cancel_warn("Incorrect picture file name" )
            return
        suff = fns[1]
        pxm = edm.view.pixmap
        pxb = Pixbuf(COLORSPACE_RGB, False, 8,
                    edm.view.draw_width, edm.view.draw_height)
        pxb.get_from_drawable(pxm, colormap_get_system(), 0, 0, 0, 0, -1, -1)
        if suff == 'png':
            pxb.save(file_name, 'png')
        elif suff == 'jpg':
            pxb.save(file_name, "jpeg", {"quality":"100"})
        else:
            cancel_warn("Unknown picture format")

    @check_curent_model
    def do_layout(self, widget, layout_style):
        """
        Compute a layout of the model
        """

        edm = self.current_edit_mvc
        lvi = LayoutVisitor(edm.view, layout_style)
        edm.model.accept(lvi)
        edm.display()

    # Simulation
    @check_curent_model
    def on_simulate(self, widget):
        """
        Call simulation
        """
        reporter = CompilReporter()
        ChartSimulControler(self.current_edit_mvc, True, reporter)

    # Static analysis
    @check_curent_model
    def on_stats_info(self, widget):
        """Static analysis: Fill window with model informations.
        """
        reporter = CompilReporter()
        # get stats from StaticAnalyzer
        stan = StaticAnalyzer(reporter)
        stan.build_from_chart_model(self.current_edit_mvc.model)
        ststat = stan.get_statistics()
        window = STATWindow(ststat, self.current_edit_mvc, reporter, self)
        # Get Main widget and connect its "destroy" event
        window.window.connect("destroy", self.win_remove)

    @check_curent_model
    def on_frontier_scc(self, widget):
        """Static analysis: Compute connected components which are on the frontier
        """
        reporter = CompilReporter()
        stan = StaticAnalyzer(reporter)
        stan.build_from_chart_model(self.current_edit_mvc.model)
        # errors??
        lscc = stan.get_frontier_scc()
        window = SCCWindow(lscc, self.current_edit_mvc, reporter, self)
        # Get Main widget and connect its "destroy" event
        window.window.connect("destroy", self.win_remove)

    @check_curent_model
    def on_basal_activated_genes(self, widget):
        """Static analysis: Compute basal activated genes
        """
        reporter = CompilReporter()
        stan = StaticAnalyzer(reporter)
        stan.build_from_chart_model(self.current_edit_mvc.model)
        lwbag = stan.get_why_basal_genes()
        window = BAGWindow(lwbag, self.current_edit_mvc, reporter, self)
        # Get Main widget and connect its "destroy" event
        window.window.connect("destroy", self.win_remove)

    @check_curent_model
    def on_dependency_graph(self, widget):
        """Static analysis: Computation and export of dependency graphs
        """
        selected_menu = widget.get_name()
        reporter = CompilReporter()
        static_analyser = StaticAnalyzer(reporter)
        static_analyser.build_from_chart_model(self.current_edit_mvc.model)

        # Detect type of selected menu
        if selected_menu == "transitionGraph":
            graph = static_analyser.make_transition_dg()
        elif selected_menu == "dependencyGraph":
            graph = static_analyser.make_dependence_dg(True)
        else:
            graph = static_analyser.make_full_dependence_dg(True)

        # Open Filechooser to export graph file
        ch_opt = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_SAVE,
                   gtk.RESPONSE_OK)
        choice = gtk.FileChooserDialog("Save the graph as ...", None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE, ch_opt)
        choice.set_default_response(gtk.RESPONSE_OK)

        choice.set_current_name(
            self.current_edit_mvc.model.name + '_' + \
            selected_menu + ".graphml")

        # add a filter to see only graphml files (*.graphml)
        filter = gtk.FileFilter()
        filter.set_name("graphml files")
        filter.add_pattern("*.graphml")
        choice.add_filter(filter)

        # add a filter to see only dot files (*.dot)
        filter = gtk.FileFilter()
        filter.set_name("dot files")
        filter.add_pattern("*.dot")
        choice.add_filter(filter)

        # add a filter to see all
        no_filter = gtk.FileFilter()
        no_filter.set_name("all")
        no_filter.add_pattern("*")
        choice.add_filter(no_filter)

        response = choice.run()

        if response == gtk.RESPONSE_OK:
            # Export to desired file type
            if 'dot' in choice.get_filter().get_name():
                # Dot
                static_analyser.export_2_dot(graph, choice.get_filename())
            else:
                # Graphml
                static_analyser.export_2_graphml(graph, choice.get_filename())

        elif response == gtk.RESPONSE_CANCEL:
            pass
        choice.destroy()

    def win_register(self, window):
        """Register a sub window
        This window will be destroyed when you connect its main widget
        event "destroy" to subwin_on_destroy()

        used by: STATWindow, SCCWindow, BAGWindow, LegendWindow
        """
        print "register"
        self.subwindows.add(window)

    def win_remove(self, window):
        """Destroy a sub window
        This window will be destroyed when you connect its main widget
        event "destroy" to subwin_on_destroy()

        used by: STATWindow, SCCWindow, BAGWindow, LegendWindow
        """
        print "remove"
        self.subwindows.remove(window)
        window.destroy()

    # Solve model
    @check_curent_model
    def on_check(self, widget):
        """
        Launch checker
        """
        reporter = CompilReporter()
        ChartChecker(self.current_edit_mvc, reporter, self)

    def get_em_with_model_name(self, name):
        """
        get em in list of edit_MVC
        """
        for edm in self.edit_mvc_list:
            if edm.title == name:
                return edm
        return None

    def update(self, mnode):
        """
        observer method
        """
        if mnode.name != self.current_edit_mvc.title:
            list_model = []
            for edm in self.edit_mvc_list:
                list_model.append(edm.model.name)
            if mnode.name in list_model:
                pass
            else:
                self.open_macro(mnode)
        else :
            pass


    def display_states(self, widget):
        """
        As it says
        """
        cst = self.chart_info.trans_info.trans.get_influencing_places()
        ndic = self.current_edit_mvc.model.simple_node_dict
        for ident in cst:
            try:
                ndic[ident].search_mark = True
            except:
                pass
        self.current_edit_mvc.view.update()

    def edit_constraints(self, widget):
        """
        Constraint editor
        """
        model = self.current_edit_mvc.model
        # what_to_do = save_constraints
        self.constraint_window =  BioSignalEditor('Constraints: '+model.name,
                                                   self, self.save_constraints)
        self.constraint_window.set_text(model.constraints)

    def save_constraints(self, edit):
        """
        For export
        """
        if self.constraint_window:
            model = self.current_edit_mvc.model
            model.constraints = edit.get_text()
            model.modified = True
            self.constraint_window = None





## Filename    : chart_chrono.py
## Author(s)   : Michel Le Borgne
## Created     : 05/2010
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##
"""
Widget for chronograms
"""
import gtk
from gtk.gdk import color_parse, Pixmap
from cadbiom_gui.gt_gui.utils.warn import cancel_warn
INTERLIGNE = 30
W_STEP = 20
H_STEP = 10
CHAR_PIX = 9

class ChartChrono (object):
    """
    A class for displaying a chronogram
    """
    def __init__(self):
        self.name = None
        self.state_names = []
        self.state_dict = dict()
        self.input_names = []
        self.input_dict = dict()
        self.left_margin = 0
        self.nb_step = 0

        self.window = None

    @classmethod
    def from_simulation_model(cls, sim):
        """
        @param sim: simulator object
        """
        chrono = cls()
        chrono.name = sim.model.name
        # initialisations
        symb_table = sim.get_symb_tab()
        lnam = 0  # max size of a name
        for key in symb_table.keys():
            elem = symb_table[key]
            if elem.is_place():
                chrono.state_names.append(key)
                lnam = max(lnam, len(key))
                chrono.state_dict[key] = []
            if elem.is_input_event():
                chrono.input_names.append(key)
                lnam = max(lnam, len(key))
                chrono.input_dict[key] = []
        chrono.state_names.sort()
        chrono.input_names.sort()
        chrono.h_extend = (len(chrono.input_names)
                           + len(chrono.state_names)) * INTERLIGNE + INTERLIGNE
        chrono.w_extend = 500 # width of the curve area
        chrono.height = min(chrono.h_extend, 1000)
        chrono.width = 500
        chrono.left_margin = lnam * CHAR_PIX
        chrono.set_view()
        return chrono

    @classmethod
    def for_trajectory(cls, sim):
        """
        version of a chronogram for registering a whole trajectory
        @param sim : the simulator
        """
        chrono = cls()
        chrono.name = sim.model.name
        # initialisations
        symb_table = sim.get_symb_tab()
        for key in symb_table.keys():
            elem = symb_table[key]
            if elem.is_place():
                chrono.state_names.append(key)
                chrono.state_dict[key] = []
            if elem.is_input_event():
                chrono.input_names.append(key)
                chrono.input_dict[key] = []
        return chrono

    def set_view(self):
        """
        establish the graphical windows
        """
        # view
        self.window =  gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.set_title("Chronogram "+self.name)
        self.window.set_position(gtk.WIN_POS_CENTER)
        self.window.set_size_request(500 + self.left_margin, self.height)

        # hbox
        hbox = gtk.HBox()
        hbox.set_homogeneous(False)
        scrw = gtk.ScrolledWindow()
        #sw.set_size_request(self.width,self.height)
        scrw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrw.set_shadow_type(gtk.SHADOW_IN)
        scrw.add_with_viewport(hbox)
        self.window.add(scrw)
        self.scrw = scrw

        # name view
        self.name_view = gtk.DrawingArea()
        self.name_view.set_size_request(self.left_margin, self.h_extend)
        self.name_pixmap = None
        self.name_view.connect("expose_event", self.__expose_event)
        self.name_view.connect("configure_event", self.__configure_event)
        hbox.pack_start(self.name_view, True, True)

        # curve view
        self.view = gtk.DrawingArea()
        #self.name_view.set_size_request(self.w_extend, self.h_extend)
        self.pixmap = None # cannot be initialized before self.window creation
        hbox.pack_start(self.view, True, False)
        self.view.connect("expose_event", self.__expose_event)
        self.view.connect("configure_event", self.__configure_event)

        adj = self.scrw.get_hadjustment()
        adj.connect("value_changed", self.__changed_event)

        self.window.show_all()

    def register(self, sim):
        """
        read the simulation model and register the state of each place
        """
        symb_table = sim.get_symb_tab()
        for key in symb_table.keys():
            elem = symb_table[key]
            if elem.is_place():
                if elem.activated:
                    self.state_dict[key].append(1)
                else:
                    self.state_dict[key].append(0)
            if elem.is_input_event():
                if elem.activated:
                    self.input_dict[key].append(1)
                else:
                    self.input_dict[key].append(0)
        self.nb_step = self.nb_step + 1

    def draw_names(self):
        """
        Write the state names
        """
        # graphic context
        pixmap = self.name_pixmap
        grc = self.name_view.window.new_gc()
        grc.set_rgb_fg_color(color_parse("black"))
        stylepango = self.name_view.create_pango_layout("")
        pixmap.draw_rectangle(self.name_view.get_style().white_gc,
                          True, 0, 0, self.left_margin+5, self.h_extend)
        polyl = [(self.left_margin -1, 0), (self.left_margin-1, self.h_extend)]
        pixmap.draw_lines(grc, polyl)
        y_line = 15   # vertical coordinate of the text line
        grc.set_rgb_fg_color(color_parse("blue"))
        for name in self.input_names:
            # put name
            stylepango.set_text(name)
            pixmap.draw_layout(grc, 2, y_line, stylepango)
            polyl = [(self.left_margin -4, y_line+16),
                     (self.left_margin, y_line+16)]
            pixmap.draw_lines(grc, polyl)
            y_line = y_line + INTERLIGNE

        grc.set_rgb_fg_color(color_parse("black"))
        for name in self.state_names:
            # put name
            stylepango.set_text(name)
            wna =  stylepango.get_pixel_size()[0]
            pixmap.draw_layout(grc, 2, y_line, stylepango)
            polyl = [(wna+2, y_line+16), (self.left_margin, y_line+16)]
            pixmap.draw_lines(grc, polyl)
            y_line = y_line + INTERLIGNE

    def draw_steps(self):
        """
        Draw the chronograms for different states
        """
        pixmap = self.pixmap
        grc = self.view.window.new_gc()
        grc.set_rgb_fg_color(color_parse("#CDFFC6"))
        nb_step = self.w_extend/W_STEP
        i = 0
        while i < nb_step:
            xch = i * W_STEP
            polyl = [(xch, 0), (xch, self.h_extend)]
            if i % 10 == 0:
                grc.set_rgb_fg_color(color_parse("green"))
            pixmap.draw_lines(grc, polyl)
            if i % 10 == 0:
                grc.set_rgb_fg_color(color_parse("#CDFFC6"))
            i = i+1

    def draw(self):
        """
        Draw all
        """
        # graphic context
        pixmap = self.pixmap
        if not pixmap:
            return
        self.view.set_size_request(self.w_extend, self.height)
        grc = self.view.window.new_gc()
        pixmap.draw_rectangle(self.view.get_style().white_gc,
                          True, 0, 0, self.w_extend, self.h_extend)
        # scale
        self.draw_steps()
        # chrono
        y_line = 15
        for name in self.input_names:
            # generate chrono
            chrono = []
            st_log = self.input_dict[name]
            x_step = 0

            for val in st_log:
                step = y_line +15 - val * H_STEP
                chrono.append((x_step, step))
                chrono.append((x_step+W_STEP, step))
                x_step = x_step + W_STEP
            if len(chrono) > 0:
                grc.set_rgb_fg_color(color_parse("blue"))
                pixmap.draw_lines(grc, chrono)
            y_line = y_line + INTERLIGNE

        for name in self.state_names:
            # generate chrono
            chrono = []
            st_log = self.state_dict[name]
            x_step = 0

            for val in st_log:
                step = y_line +15 - val * H_STEP
                chrono.append((x_step, step))
                chrono.append((x_step+W_STEP, step))
                x_step = x_step + W_STEP
            if len(chrono)>0:
                grc.set_rgb_fg_color(color_parse("red"))
                pixmap.draw_lines(grc, chrono)
            y_line = y_line + INTERLIGNE

    def update(self):
        """
        Refresh
        """
        if not self.window or not self.pixmap:
            return
        cwidth = self.nb_step*W_STEP + 10
        if cwidth > self.w_extend:
            self.w_extend = cwidth
            self.pixmap = Pixmap(self.view.window, self.w_extend, self.h_extend)
            adj = self.scrw.get_hadjustment()
            adj.set_value(adj.upper)
        self.draw()
        self.view.set_size_request(self.w_extend, self.h_extend)
        fcgc = self.view.get_style().fg_gc[gtk.STATE_NORMAL]
        self.view.window.draw_drawable(fcgc,
                                       self.pixmap, 0, 0, 0, 0,
                                       self.w_extend, self.h_extend)

    def rewind(self):
        """
        As it says
        """
        for key in self.state_names:
            self.state_dict[key] = []
        for key in self.input_names:
            self.input_dict[key] = []
        self.nb_step = 0
        self.width = 500
        adj = self.scrw.get_hadjustment()
        if adj:
            adj.set_value(adj.lower)

    def select_and_prepare(self):
        """
        Select the places with a change in the activation state
        Prepare the chronogram to be displayed
        """
        old_state_dict = self.state_dict
        old_input_dict = self.input_dict
        self.state_dict = dict()
        self.input_dict = dict()
        old_state_names = self.state_names
        old_input_names = self.input_names
        self.state_names = []
        self.input_names = []
        lnam = 0
        for key in old_state_names:
            if has_changed(old_state_dict[key]):
                self.state_names.append(key)
                self.state_dict[key] = old_state_dict[key]
                lnam = max(lnam, len(key))

        for key in old_input_names:
            if has_changed(old_input_dict[key]):
                self.input_names.append(key)
                self.input_dict[key] = old_input_dict[key]
                lnam = max(lnam, len(key))
        self.state_names.sort()
        self.input_names.sort()

        if lnam == 0:
            cancel_warn("No change in the system!!")
            return False
        n_hextend = len(self.input_names)
        n_hextend = (n_hextend + len(self.state_names)) * INTERLIGNE
        self.h_extend = n_hextend + INTERLIGNE

        self.w_extend = 500
        self.height = min(self.h_extend, 1000)
        self.width = self.w_extend
        self.left_margin = lnam * CHAR_PIX
        return True

    def display_selected_chrono(self):
        """
        As it says
        """
        if self.select_and_prepare():
            self.set_view()
            self.draw()

    # window management
    def destroy(self):
        """
        Used when windows are linked
        """
        if self.window:
            self.window.destroy()

    def on_destroy(self):
        """
        Standard call back
        """
#        if self.parent:
#            self.parent.win_remove(self)
        if self.window:
            self.window.destroy()

    def __expose_event(self, widget, event):
        """
        Redisplay drawing after cover and exposure of the window
        """
        if not self.view.window:
            return True
        if not self.name_pixmap:
            self.name_pixmap = Pixmap(self.name_view.window,
                                      self.left_margin, self.h_extend)
            self.draw_names()
        if not self.pixmap:
            self.pixmap = Pixmap(self.view.window,
                                 self.w_extend, self.h_extend)
            self.draw()
        # expose the drawing  already in the pixmaps
        style = widget.get_style().fg_gc[gtk.STATE_NORMAL]
        self.name_view.window.draw_drawable(style, self.name_pixmap,
                                            0, 0, 0, 0,
                                            self.left_margin, self.h_extend)
        self.view.window.draw_drawable(style, self.pixmap,
                                       0, 0, 0, 0, self.w_extend, self.h_extend)
        return True

    def __configure_event(self, widget, event):
        """
        Initialisation
        """
        if not self.name_pixmap:
            self.name_pixmap = Pixmap(self.name_view.window,
                                       self.left_margin, self.h_extend)
        self.draw_names()
        if not self.pixmap:
            #self.pixmap = Pixmap(self.view.window,
            #                     self.w_extend, self.h_extend)
            pass
        else:
            self.draw()
        return True

    def __changed_event(self, widget):
        """
        change in size
        """
        style = self.view.get_style().fg_gc[gtk.STATE_NORMAL]
        self.view.window.draw_drawable(style, self.pixmap,
                                       0, 0, 0, 0,
                                       self.w_extend, self.h_extend)
        return True


# helper function
def has_changed(lls):
    """
    Test changing from last call
    """
    xxx = lls[0]
    for elem in lls:
        if not xxx == elem:
            return True
    return False



